from __future__ import annotations

from abc import ABC
from dataclasses import dataclass, field
from typing import MutableSequence

from aeon.core.types import (
    AbstractionType,
    BaseKind,
    RefinedType,
    Top,
    TypeConstructor,
    TypePolymorphism,
    TypeVar,
    builtin_core_types,
)
from aeon.core.types import Kind
from aeon.core.types import StarKind
from aeon.core.types import Type
from aeon.utils.name import Name


class TypingContextEntry(ABC):
    pass


@dataclass(unsafe_hash=True)
class VariableBinder(TypingContextEntry):
    name: Name
    type: Type

    def __repr__(self):
        return f"{self.name} : {self.type}"


@dataclass(unsafe_hash=True)
class UninterpretedBinder(TypingContextEntry):
    name: Name
    type: AbstractionType

    def __repr__(self):
        return f"uninterpreted {self.name} : {self.type}"


@dataclass(unsafe_hash=True)
class TypeBinder(TypingContextEntry):
    type_name: Name
    type_kind: Kind = field(default_factory=StarKind)

    def __repr__(self):
        return f"type {self.type_name} {self.type_kind}"


@dataclass
class TypeConstructorBinder(TypingContextEntry):
    name: Name
    args: list[Name]  # cant hash

    def __repr__(self):
        if self.args:
            argsf = "(" + ", ".join(map(str, self.args)) + ")"
        else:
            argsf = ""
        return f"type {self.name}{argsf}"

    def __hash__(self):
        return hash(self.name) + hash(tuple(self.args))


@dataclass
class TypingContext:
    entries: MutableSequence[TypingContextEntry] = field(default_factory=list)

    def __post_init__(self):
        for bt in builtin_core_types[::-1]:
            temp = TypeConstructorBinder(bt.name, [])
            if temp not in self.entries:
                self.entries.insert(0, temp)

    def __repr__(self):
        fields = "; ".join(map(repr, self.entries))
        return f"[[{fields}]]"

    def with_var(self, name: Name, type: Type) -> TypingContext:
        nentries = [e for e in self.entries] + [VariableBinder(name, type)]
        return TypingContext(nentries)

    def with_typevar(self, name: Name, kind: Kind) -> TypingContext:
        nentries = [e for e in self.entries] + [TypeBinder(name, kind)]
        return TypingContext(nentries)

    def type_of(self, name: Name) -> Type | None:
        for e in self.entries:
            match e:
                case VariableBinder(iname, ty):
                    if iname == name:
                        return ty
        return None

    def kind_of(self, ty: Type) -> Kind:
        match ty:
            case Top() | RefinedType(_, TypeConstructor(_), _) | RefinedType(_, TypeConstructor(_, _), _):
                return BaseKind()
            case TypeVar(name):
                assert (name, BaseKind()) in self.typevars()
                # TODO Polytypes: What it * is in context?
                return BaseKind()
            case RefinedType(_, TypeVar(name), _):
                assert (name, BaseKind()) in self.typevars(), f"{name} not in {self.typevars()}"
                return BaseKind()
            case AbstractionType(_, _, _):
                return StarKind()
            case TypePolymorphism(_, _, _):
                return StarKind()
            case TypeConstructor(_, _):
                return BaseKind()
            case _:
                assert False, f"Unknown type in context: {ty}"

    def typevars(self) -> list[tuple[Name, Kind]]:
        return [(e.type_name, e.type_kind) for e in self.entries if isinstance(e, TypeBinder)]

    def vars(self) -> list[tuple[Name, Type]]:
        return [
            (e.name, e.type)
            for e in self.entries
            if isinstance(e, VariableBinder) or isinstance(e, UninterpretedBinder)
        ]

    def concrete_vars(self) -> list[tuple[Name, Type]]:
        return [(e.name, e.type) for e in self.entries if isinstance(e, VariableBinder)]

    def has_uninterpreted_fun(self, name) -> bool:
        return name in [e.name for e in self.entries if isinstance(e, UninterpretedBinder)]

    def get_type_constructor(self, name: Name) -> list[Name] | None:
        for entry in self.entries[::-1]:
            match entry:
                case TypeConstructorBinder(ename, eargs):
                    if ename == name:
                        return eargs
        # TODO: enter in context.
        if name.name in ["Unit", "Int", "Bool", "Float", "String"]:
            return []
        return None

    def __hash__(self):
        return sum(hash(e) for e in self.entries)



from typing import Any, Callable

from aeon.utils.name import Name
from aeon.sugar.program import Definition, STerm

Metadata = dict[Name, Any]
DecoratorType = Callable[[list[STerm], Definition, Metadata], tuple[Definition, list[Definition], Metadata]]


def metadata_update(metadata: Metadata, fun: Definition, aux_dict: dict[str, Any] = None) -> Metadata:
    if not aux_dict:
        aux_dict = {}
    if fun.name in metadata.keys():
        metadata[fun.name].update(aux_dict)
    else:
        metadata[fun.name] = aux_dict
    return metadata



from aeon.synthesis.api import Synthesizer
from aeon.synthesis.grammar.ge_synthesis import GESynthesizer
from aeon.synthesis.modules.synquid.synthesizer import SynquidSynthesizer
from aeon.synthesis.modules.llm import LLMSynthesizer


def make_synthesizer(module: str) -> Synthesizer:
    match module:
        case "random_search":
            return GESynthesizer(method="random_search")
        case "enumerative":
            return GESynthesizer(method="enumerative")
        case "gp":
            return GESynthesizer(method="genetic_programming")
        case "1p1":
            return GESynthesizer(method="one_plus_one")
        case "hc":
            return GESynthesizer(method="hill_climbing")
        case "synquid":
            return SynquidSynthesizer()
        case "llm":
            return LLMSynthesizer()
        case _:
            assert False, f"Not supported synthesizer with name {module}"



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field

from aeon.utils.name import Name
from aeon.core.types import Kind
from aeon.sugar.stypes import SType, STypeConstructor

from aeon.utils.location import Location, SynthesizedLocation


class STerm:
    loc: Location

    def __hash__(self) -> int:
        return str(self).__hash__()


@dataclass(frozen=True)
class SLiteral(STerm):
    value: object
    type: SType
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __repr__(self):
        if type(self.value) is str:
            return f'"{self.value}"'
        return f"{self.value}".lower()

    def __eq__(self, other):
        return isinstance(other, SLiteral) and self.value == other.value and self.type == other.type

    def __str__(self):
        if self.type == STypeConstructor(Name("String", 0)):
            return f'"{self.value}"'
        return f"{self.value}"


@dataclass(frozen=True)
class SVar(STerm):
    name: Name
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"{self.name}"

    def __repr__(self):
        return f"Var({self.name})"

    def __eq__(self, other):
        return isinstance(other, SVar) and self.name == other.name


@dataclass(frozen=True)
class SAnnotation(STerm):
    expr: STerm
    type: SType
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"({self.expr} : {self.type})"

    def __repr__(self):
        return f"({self.expr} : {self.type})"

    def __eq__(self, other):
        return isinstance(other, SAnnotation) and self.expr == other.expr


@dataclass(frozen=True)
class SHole(STerm):
    name: Name
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"?{self.name}"

    def __repr__(self):
        return f"Hole({self.name})"

    def __eq__(self, other):
        return isinstance(other, SHole) and self.name == other.name


@dataclass(frozen=True)
class SApplication(STerm):
    fun: STerm
    arg: STerm
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __repr__(self):
        return f"({self.fun} {self.arg})"

    def __eq__(self, other):
        return isinstance(other, SApplication) and self.fun == other.fun and self.arg == other.arg

    def __str__(self):
        return f"({self.fun} {self.arg})"


@dataclass(frozen=True)
class SAbstraction(STerm):
    var_name: Name
    body: STerm
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"(\\{self.var_name} -> {self.body})"

    def __eq__(self, other):
        return isinstance(other, SAbstraction) and self.var_name == other.var_name and self.body == other.body


@dataclass(frozen=True)
class SLet(STerm):
    var_name: Name
    var_value: STerm
    body: STerm
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"(let {self.var_name} = {self.var_value} in\n\t{self.body})"

    def __eq__(self, other):
        return (
            isinstance(other, SLet)
            and self.var_name == other.var_name
            and self.var_value == other.var_value
            and self.body == other.body
        )


@dataclass(frozen=True)
class SRec(STerm):
    var_name: Name
    var_type: SType
    var_value: STerm
    body: STerm
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "(let {} : {} = {} in\n\t{})".format(
            self.var_name,
            self.var_type,
            self.var_value,
            self.body,
        )

    def __eq__(self, other):
        return (
            isinstance(other, SRec)
            and self.var_name == other.var_name
            and self.var_type == other.var_type
            and self.var_value == other.var_value
            and self.body == other.body
        )


@dataclass(frozen=True)
class SIf(STerm):
    cond: STerm
    then: STerm
    otherwise: STerm
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"(if {self.cond} then {self.then} else {self.otherwise})"

    def __eq__(self, other):
        return (
            isinstance(other, SIf)
            and self.cond == other.cond
            and self.then == other.then
            and self.otherwise == other.otherwise
        )


@dataclass(frozen=True)
class STypeAbstraction(STerm):
    name: Name
    kind: Kind
    body: STerm
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"ƛ{self.name}:{self.kind}.({self.body})"


@dataclass()  # frozen=True
class STypeApplication(STerm):
    body: STerm
    type: SType
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"({self.body})[{self.type}]"


class Node:
    pass


@dataclass
class ImportAe(Node):
    path: str
    func: str
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        if not self.func:
            return f"import {self.path};"
        else:
            return f"import {self.func} from {self.path};"


@dataclass
class TypeDecl(Node):
    name: Name
    args: list[Name] = field(default_factory=list)
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"type {self.name};"


@dataclass
class InductiveDecl(Node):
    name: Name
    args: list[Name] = field(default_factory=list)
    constructors: list[Definition] = field(default_factory=list)
    measures: list[Definition] = field(default_factory=list)
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __post_init__(self):
        assert isinstance(self.name, Name)

        for aname in self.args:
            assert isinstance(aname, Name)

    def __str__(self):
        args = " ".join(str(arg) for arg in self.args)
        constructors = " ".join(f"| {cons}" for (cons) in self.constructors)
        measures = " ".join(f"+ {dec}" for dec in self.measures)
        return f"inductive {self.name} {args} {constructors} {measures}"


@dataclass
class Decorator(Node):
    name: Name
    macro_args: list[STerm]
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __repr__(self):
        macro_args = ", ".join([f"{term}" for (term) in self.macro_args])
        return f"@{self.name}({macro_args})"


@dataclass
class Definition(Node):
    name: Name
    foralls: list[tuple[Name, Kind]]
    args: list[tuple[Name, SType]]
    type: SType
    body: STerm
    decorators: list[Decorator] = field(default_factory=list)
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __post_init__(self):
        assert isinstance(self.type, SType)

    def __str__(self):
        if not self.args:
            return f"def {self.name} : {self.type} = {self.body};"
        else:
            args = ", ".join([f"{n}:{t}" for (n, t) in self.args])
            foralls = " ".join([f"∀{n}:{k}" for (n, k) in self.foralls])
            return f"def {self.name} {foralls} {args} : {self.type} {{\n {self.body} \n}}"


@dataclass
class Program(Node):
    imports: list[ImportAe]
    type_decls: list[TypeDecl]
    inductive_decls: list[InductiveDecl]
    definitions: list[Definition]

    def __str__(self):
        imps = "\n".join([str(td) for td in self.imports])
        decls = "\n".join([str(td) for td in self.type_decls])
        inductives = "\n".join([str(td) for td in self.inductive_decls])
        defs = "\n".join([str(d) for d in self.definitions])
        return f"{imps}\n{decls}\n{inductives}\n{defs}"



from typing import Any
from aeon.backend.evaluator import EvaluationContext
from aeon.core.terms import Term
from aeon.core.terms import Hole
from aeon.core.types import Type
from aeon.utils.name import Name
from aeon.synthesis.uis.api import SynthesisUI
from aeon.typechecking.context import TypingContext


class TerminalUI(SynthesisUI):
    best_solution: Term
    best_quality: list[float] | None

    def start(
        self,
        typing_ctx: TypingContext,
        evaluation_ctx: EvaluationContext,
        target_name: str,
        target_type: Type,
        budget: Any,
    ):
        self.target_name = target_name
        self.target_type = target_type
        self.budget = budget
        self.best_solution = Hole(Name("sorry", -1))
        self.best_quality = None

    def register(
        self,
        solution: Term,
        quality: Any,
        elapsed_time: float,
        is_best: bool,
    ):
        if is_best:
            self.best_solution = solution
            self.best_quality = quality

        print(
            f"Target: {self.target_name} ({elapsed_time:.1f} / {self.budget:.1f}s) "
            + f"| Best: {self.best_solution} ({self.best_quality}) "
            + f"| Current: {solution} ({quality})",
        )

    def end(self, solution: Term, quality: Any):
        pass



from aeon.core.liquid import (
    LiquidApp,
    LiquidLiteralBool,
    LiquidLiteralFloat,
    LiquidLiteralInt,
    LiquidLiteralString,
    LiquidTerm,
    LiquidVar,
)
from aeon.core.types import LiquidHornApplication, TypeConstructor, t_int
from aeon.core.substitutions import substitution_in_liquid, substitute_vartype
from aeon.core.terms import (
    Abstraction,
    Annotation,
    Application,
    If,
    Let,
    Literal,
    Rec,
    Term,
    TypeAbstraction,
    TypeApplication,
    Var,
    Hole,
)
from aeon.core.types import AbstractionType, RefinedType, Type, TypePolymorphism, Top, TypeVar, t_unit
from aeon.elaboration.context import (
    ElabTypeDecl,
    ElabTypeVarBinder,
    ElabTypingContextEntry,
    ElabUninterpretedBinder,
    ElabVariableBinder,
    ElaborationTypingContext,
)
from aeon.sugar.program import (
    SAbstraction,
    SAnnotation,
    SApplication,
    SIf,
    SLet,
    SLiteral,
    SRec,
    STerm,
    STypeAbstraction,
    STypeApplication,
    SVar,
    SHole,
)
from aeon.sugar.stypes import (
    SAbstractionType,
    SRefinedType,
    SType,
    STypeConstructor,
    STypePolymorphism,
    STypeVar,
)
from aeon.sugar.substitutions import normalize, substitution_sterm_in_sterm, substitution_sterm_in_stype
from aeon.typechecking.context import (
    TypeBinder,
    TypeConstructorBinder,
    TypingContext,
    TypingContextEntry,
    UninterpretedBinder,
    VariableBinder,
)
from aeon.utils.name import Name, fresh_counter


class LoweringException(Exception):
    pass


class LiquefactionException(LoweringException):
    pass


# TODO: NOW! detect built-in SMT functions
def liquefy_app(app: SApplication) -> LiquidApp | None:
    arg = liquefy(app.arg)
    fun = app.fun
    while isinstance(fun, STypeApplication):
        fun = fun.body
    if not arg:
        return None

    match fun:
        case SVar(name):
            return LiquidApp(name, [arg], loc=app.loc)
        case SApplication(_, _):
            liquid_pseudo_fun = liquefy_app(fun)
            if liquid_pseudo_fun:
                return LiquidApp(liquid_pseudo_fun.fun, liquid_pseudo_fun.args + [arg], loc=app.loc)
            return None
        case SLet(name, val, body):
            app = SApplication(substitution_sterm_in_sterm(body, val, name), app.arg, loc=app.loc)
            return liquefy_app(app)
        case _:
            raise LiquefactionException(f"{app} is not a valid predicate.")


def liquefy(t: STerm, available_vars: list[tuple[Name, TypeConstructor | TypeVar]] | None = None) -> LiquidTerm:
    """Converts Surface Terms into Liquid Terms"""
    match t:
        case SLiteral(val, STypeConstructor(Name("Bool", _)), loc):
            assert isinstance(val, bool)
            return LiquidLiteralBool(val, loc=loc)
        case SLiteral(val, STypeConstructor(Name("Int", _)), loc):
            assert isinstance(val, int)
            return LiquidLiteralInt(val, loc=loc)
        case SLiteral(val, STypeConstructor(Name("Float", _)), loc):
            assert isinstance(val, float)
            return LiquidLiteralFloat(val, loc=loc)
        case SLiteral(val, STypeConstructor(Name("String", _)), loc):
            assert isinstance(val, str)
            return LiquidLiteralString(val, loc=loc)
        case SLiteral(_, _):
            assert False, f"{t} is not convertable to liquid term."
        case SVar(name, loc):
            return LiquidVar(name, loc=loc)
        case SIf(cond, then, otherwise, loc):
            co = liquefy(cond, available_vars)
            th = liquefy(then, available_vars)
            ot = liquefy(otherwise, available_vars)
            if co is not None and th is not None and ot is not None:
                return LiquidApp(Name("ite", 0), [co, th, ot], loc=loc)
            return None
        case SAnnotation(expr, _):
            return liquefy(expr, available_vars)
        case SAbstraction(name, body):
            return None
        case STypeApplication(expr, _):
            return liquefy(expr, available_vars)
        case STypeAbstraction(name, _, body):
            return liquefy(body, available_vars)
        case SApplication(_, _):
            return liquefy_app(t)
        case SLet(name, val, body):
            lval = liquefy(val, available_vars)
            lbody = liquefy(body, available_vars)
            if lval and lbody:
                return substitution_in_liquid(lbody, lval, name)
            return None
        case SRec(name, _, val, body):
            lval = liquefy(val, available_vars)  # TODO: induction?
            lbody = liquefy(body, available_vars)
            if lval and lbody:
                return substitution_in_liquid(lbody, lval, name)
            return None
        case SHole(name, loc):
            avars = available_vars or []
            return LiquidHornApplication(name=name, argtypes=[(LiquidVar(x), ty) for (x, ty) in avars], loc=loc)
        case _:
            assert False


def basic_type(ty: Type) -> TypeConstructor | TypeVar:
    match ty:
        case TypeConstructor(_, _) | TypeVar(_):
            return ty
        case RefinedType(_, it, _):
            return basic_type(it)
        case Top():
            return t_unit
        case _:
            assert False, f"Unknown base type {ty} ({type(ty)})"


def type_to_core(ty: SType, available_vars: list[tuple[Name, TypeConstructor | TypeVar]] | None = None) -> Type:
    """Converts Surface Types into Core Types"""

    if available_vars is None:
        available_vars = []

    match normalize(ty):
        case STypeConstructor(Name("Top", 0), loc):
            return Top()  # TODO: loc?
        case STypeVar(name, loc):
            return TypeVar(name, loc=loc)
        case SAbstractionType(name, vty, rty, loc):
            nname = Name(name.name, fresh_counter.fresh())
            at = type_to_core(vty, available_vars)
            if isinstance(at, TypeConstructor) or isinstance(at, TypeVar) or isinstance(at, RefinedType):
                available_vars = available_vars + [(nname, basic_type(at))]
                nrty = substitution_sterm_in_stype(rty, SVar(nname), name)
            else:
                nrty = rty
            return AbstractionType(nname, at, type_to_core(nrty, available_vars), loc=loc)
        case STypePolymorphism(name, kind, rty, loc):
            return TypePolymorphism(name, kind, type_to_core(rty, available_vars), loc=loc)
        case SRefinedType(oname, ity, ref, loc):
            if oname.id == -1:
                name = Name(oname.name, fresh_counter.fresh())
                ref = substitution_sterm_in_sterm(ref, SVar(name), oname)
            else:
                name = oname
            basety = type_to_core(ity, available_vars)
            assert (
                isinstance(basety, TypeConstructor)
                or isinstance(basety, TypeVar)
                or isinstance(basety, TypeConstructor)
            )
            return RefinedType(name, basety, liquefy(ref, available_vars + [(name, basety)]), loc=loc)
        case STypeConstructor(name, args, loc):
            return TypeConstructor(name, [type_to_core(ity, available_vars) for ity in args], loc=loc)
        case _:
            assert False, f"Unknown {ty} / {normalize(ty)}."


def lower_to_core(t: STerm) -> Term:
    """Converts Surface terms into Core terms."""
    match t:
        case SHole(name, loc):
            return Hole(name, loc=loc)
        case SLiteral(val, ty, loc):
            return Literal(val, type_to_core(ty), loc=loc)
        case SVar(name, loc):
            return Var(name, loc=loc)
        case SIf(cond, then, otherwise, loc):
            return If(lower_to_core(cond), lower_to_core(then), lower_to_core(otherwise), loc=loc)
        case SApplication(fun, arg, loc):
            return Application(lower_to_core(fun), lower_to_core(arg), loc=loc)
        case SLet(name, val, body, loc):
            return Let(name, lower_to_core(val), lower_to_core(body), loc=loc)
        case SRec(name, ty, val, body, loc):
            return Rec(name, type_to_core(ty), lower_to_core(val), lower_to_core(body), loc=loc)
        case SAnnotation(expr, ty, loc):
            return Annotation(lower_to_core(expr), type_to_core(ty), loc=loc)
        case SAbstraction(name, body, loc):
            return Abstraction(name, lower_to_core(body), loc=loc)
        case STypeApplication(expr, ty, loc):
            return TypeApplication(lower_to_core(expr), type_to_core(ty), loc=loc)
        case STypeAbstraction(name, kind, body, loc):
            return TypeAbstraction(name, kind, lower_to_core(body), loc=loc)
        case _:
            assert False, f"{t} ({type(t)}) not supported"


def monomorphic_type(ty: Type) -> AbstractionType:
    match ty:
        case TypePolymorphism(name, _, body):
            return monomorphic_type(substitute_vartype(body, t_int, name))
        case AbstractionType(_, _, _):
            return ty
        case _:
            assert False, f"Type {ty} is not a monomorphic type, cannot be used in uninterpreted binders."


def wrap_ctx_entry(e: ElabTypingContextEntry) -> TypingContextEntry:
    match e:
        case ElabVariableBinder(name, ty):
            return VariableBinder(name, type_to_core(ty))
        case ElabUninterpretedBinder(name, ty):
            absty = type_to_core(ty)
            concrete_absty = monomorphic_type(absty)
            return UninterpretedBinder(name, concrete_absty)
        case ElabTypeVarBinder(name, kind):
            return TypeBinder(name, kind)
        case ElabTypeDecl(name, args):
            return TypeConstructorBinder(name, args)
        case _:
            assert False, f"{e} not supported in Core."


def lower_to_core_context(elctx: ElaborationTypingContext) -> TypingContext:
    """Lowers the elaboration context down to the Core Typing Context."""
    return TypingContext([wrap_ctx_entry(e) for e in elctx.entries])



from aeon.core.substitutions import substitute_vartype, substitution_in_type
from aeon.core.terms import (
    Abstraction,
    Annotation,
    Application,
    Hole,
    If,
    Let,
    Literal,
    Rec,
    Term,
    TypeAbstraction,
    TypeApplication,
    Var,
)
from aeon.core.types import AbstractionType, TypePolymorphism, TypeVar, refined_to_unrefined_type
from aeon.core.types import Type
from aeon.typechecking.context import TypingContext
from aeon.typechecking.typeinfer import synth
from aeon.utils.name import Name


# dict (hole_name , (hole_type, hole_typingContext))
def get_holes_info(
    ctx: TypingContext,
    t: Term,
    ty: Type,
    targets: list[tuple[Name, list[Name]]],
    refined_types: bool = False,
) -> dict[Name, tuple[Type, TypingContext]]:
    """Retrieve the Types of "holes" in a given Term and TypingContext.

    This function recursively navigates through the Term 't', updating the TypingContext and hole Type as necessary.
    When a hole is found, its Type and the current TypingContext are added to a dictionary, with the hole name as key.

    Args:
        ctx (TypingContext): The current TypingContext.
        t (Term): The term to analyze.
        ty (Type): The current type.
        targets (list(tuple(str, list(str)))): List of tuples functions names that contains holes and the name holes
        refined_types (bool): Whether to use refined types.
    """
    # ty = ty if refined_types else refined_to_unrefined_type(ty)
    match t:
        case Annotation(expr=Hole(name=hname), type=hty):
            hty = hty if refined_types else refined_to_unrefined_type(hty)
            return {hname: (hty, ctx)} if hname != "main" else {}
        case Hole(name=hname):
            return {hname: (ty, ctx)} if hname != "main" else {}
        case Literal(_, _):
            return {}
        case Var(_):
            return {}
        case Annotation(expr=expr, type=ty):
            ty = ty if refined_types else refined_to_unrefined_type(ty)
            return get_holes_info(ctx, expr, ty, targets, refined_types)
        case Application(fun=fun, arg=arg):
            hs1 = get_holes_info(ctx, fun, ty, targets, refined_types)
            hs2 = get_holes_info(ctx, arg, ty, targets, refined_types)
            return hs1 | hs2
        case If(cond=cond, then=then, otherwise=otherwise):
            hs1 = get_holes_info(ctx, cond, ty, targets, refined_types)
            hs2 = get_holes_info(ctx, then, ty, targets, refined_types)
            hs3 = get_holes_info(ctx, otherwise, ty, targets, refined_types)
            return hs1 | hs2 | hs3
        case Abstraction(var_name=vname, body=body):
            if isinstance(ty, AbstractionType):
                ret = substitution_in_type(ty.type, Var(vname), ty.var_name)
                ctx = ctx.with_var(vname, ty.var_type)
                return get_holes_info(ctx, body, ret, targets, refined_types)
            else:
                assert False, f"Synthesis cannot infer the type of {t} with type {ty}"
        case Let(var_name=vname, var_value=value, body=body):
            _, t1 = synth(ctx, value)
            t1 = t1 if refined_types else refined_to_unrefined_type(t1)
            if not isinstance(value, Hole) and not (isinstance(value, Annotation) and isinstance(value.expr, Hole)):
                ctx = ctx.with_var(vname, t1)
                hs1 = get_holes_info(ctx, t.var_value, ty, targets, refined_types)
                hs2 = get_holes_info(ctx, t.body, ty, targets, refined_types)
            else:
                hs1 = get_holes_info(ctx, t.var_value, ty, targets, refined_types)
                ctx = ctx.with_var(vname, t1)
                hs2 = get_holes_info(ctx, t.body, ty, targets, refined_types)
            return hs1 | hs2
        case Rec(var_name=vname, var_type=vtype, var_value=value, body=body):
            vtype = vtype if refined_types else refined_to_unrefined_type(vtype)
            if isinstance(vtype, AbstractionType) or isinstance(vtype, TypePolymorphism):
                hs1 = get_holes_info(ctx.with_var(vname, vtype), value, vtype, targets, refined_types)
            else:
                hs1 = get_holes_info(ctx, value, vtype, targets, refined_types)
            hs2 = get_holes_info(ctx.with_var(vname, vtype), body, ty, targets, refined_types)
            return hs1 | hs2
        case TypeApplication(body=body, type=argty):
            _, bty = synth(ctx, body)
            argty = argty if refined_types else refined_to_unrefined_type(argty)
            if isinstance(bty, TypePolymorphism):
                ntype = substitute_vartype(bty.body, argty, bty.name)
                ntype = ntype if refined_types else refined_to_unrefined_type(ntype)
                return get_holes_info(ctx, body, ntype, targets, refined_types)
            else:
                assert False, f"Synthesis cannot infer the type of {t}"
        case TypeAbstraction(name=n, kind=k, body=body):
            match ty:
                case TypePolymorphism(n2, k2, ity):
                    assert k == k2, "Kinds do not match"
                    return get_holes_info(
                        ctx.with_typevar(n, k), body, substitute_vartype(ity, TypeVar(n), n2), targets, refined_types
                    )
                case _:
                    assert False, "TypeAbstraction does not have the TypePolymorphism type."

        case _:
            assert False, f"Could not infer the type of {t} for synthesis."


def get_holes(term: Term) -> list[Name]:
    """Returns the names of holes in a particular term."""
    match term:
        case Hole(name=name):
            return [name]
        case Hole(name=name):
            return [name]
        case Literal(_):
            return []
        case Var(_):
            return []
        case Annotation(expr=expr, type=_):
            return get_holes(expr)
        case Application(fun=fun, arg=arg):
            return get_holes(fun) + get_holes(arg)
        case If(cond=cond, then=then, otherwise=otherwise):
            return get_holes(cond) + get_holes(then) + get_holes(otherwise)
        case Abstraction(var_name=_, body=body):
            return get_holes(body)
        case Let(var_name=_, var_value=value, body=body):
            return get_holes(value) + get_holes(body)
        case Rec(var_name=_, var_type=_, var_value=value, body=body):
            return get_holes(value) + get_holes(body)
        case TypeApplication(body=body, type=_):
            return get_holes(body)
        case TypeAbstraction(name=_, kind=_, body=body):
            return get_holes(body)
        case _:
            assert False


def iterate_top_level(term: Term):
    """Iterates through a program, and returns the top-level functions."""
    while isinstance(term, Rec):
        yield term
        term = term.body


def incomplete_functions_and_holes(ctx: TypingContext, term: Term) -> list[tuple[Name, list[Name]]]:
    """Given a typing context and a term, this function identifies which top-
    level functions have holes, and returns a list of holes in each
    function."""
    return [(rec.var_name, get_holes(rec.var_value)) for rec in iterate_top_level(term) if get_holes(rec.var_value)]



from __future__ import annotations

from aeon.core.liquid import LiquidLiteralBool
from aeon.core.types import AbstractionType, BaseKind, Top, TypeConstructor
from aeon.core.types import Kind
from aeon.core.types import RefinedType
from aeon.core.types import StarKind
from aeon.core.types import t_bool
from aeon.core.types import Type
from aeon.core.types import TypePolymorphism
from aeon.core.types import TypeVar
from aeon.typechecking.context import TypingContext
from aeon.typechecking.liquid import typecheck_liquid


def wf_inner(ctx: TypingContext, t: Type, k: Kind = StarKind()) -> bool:
    match t:
        case Top():
            return True  # wf_no_refinement
        case RefinedType(name, TypeConstructor(_, _) as ty, refinement):
            inferred_type = typecheck_liquid(ctx.with_var(name, ty), refinement)
            return inferred_type == t_bool
        case TypeVar(tvname):
            return tvname in [v[0] for v in ctx.typevars()]
        case RefinedType(name, TypeVar(tvname), LiquidLiteralBool(True)):
            return (tvname, k) in ctx.typevars()
        case RefinedType(name, TypeVar(tvname) as ty, refinement):
            return (
                k == BaseKind()
                and (tvname, BaseKind()) in ctx.typevars()
                and typecheck_liquid(ctx.with_var(name, ty), refinement) == t_bool
            )
        case AbstractionType(aname, atype, rtype):
            return k == StarKind() and wellformed(ctx, atype) and wellformed(ctx.with_var(aname, atype), rtype)
        case TypePolymorphism(name, kind, body):
            return k == StarKind() and wellformed(ctx.with_typevar(name, kind), body)
        case TypeConstructor(name, args):
            if not args:
                return ctx.get_type_constructor(name) is not None
            else:
                cargs = ctx.get_type_constructor(name)
                if not cargs or len(cargs) != len(args):
                    return False
                return all(wf_inner(ctx, t) for t in args)
        case RefinedType(name, TypeConstructor(_, args) as ity, refinement):
            return wf_inner(ctx, ity) and typecheck_liquid(ctx.with_var(name, ity), refinement) == t_bool
        case _:
            return False


def wellformed(ctx: TypingContext, t: Type, k: Kind = StarKind()) -> bool:
    match k:
        case StarKind():
            return wf_inner(ctx, t, StarKind()) or wf_inner(ctx, t, BaseKind())
        case _:
            return wf_inner(ctx, t, BaseKind())



from __future__ import annotations

from functools import wraps
from time import process_time
from time import perf_counter

from loguru import logger


def measure(func):
    @wraps(func)
    def _time_it(*args, **kwargs):
        start = int(round(process_time() * 1000))
        try:
            return func(*args, **kwargs)
        finally:
            end_ = int(round(process_time() * 1000)) - start
            if end_ > 1:
                end_s = end_ if end_ > 0 else 0
                logger.info(f"Total execution time {func.__name__}: {end_s} ms")

    return _time_it


class RecordTime:
    def __init__(self, name: str):
        self.name = name

    def __enter__(self):
        self.start = perf_counter()
        return self

    def __exit__(self, type, value, traceback):
        self.time = perf_counter() - self.start
        logger.log("TIME", f"[{self.name}] Time: {self.time:.3f} seconds")



from aeon.core.liquid import (
    LiquidApp,
    LiquidLiteralBool,
    LiquidLiteralFloat,
    LiquidLiteralInt,
    LiquidLiteralString,
    LiquidTerm,
    LiquidVar,
)
from aeon.utils.name import Name
from aeon.core.terms import Term, Literal, Var, Application, Abstraction, Let, Rec, If, TypeAbstraction, TypeApplication
from aeon.core.types import (
    LiquidHornApplication,
    Type,
    TypeVar,
    AbstractionType,
    RefinedType,
    TypePolymorphism,
    TypeConstructor,
    Top,
)


def core_liquid_equality(lq1: LiquidTerm, lq2: LiquidTerm, rename_left: dict[Name, Name] | None = None) -> bool:
    "Equality of liquid terms up to alpha renaming"
    rename_left = rename_left or {}
    match lq1, lq2:
        case LiquidVar(an), LiquidVar(bn):
            return rename_left.get(an, an) == bn
        case LiquidLiteralBool(a), LiquidLiteralBool(b):
            return a == b
        case LiquidLiteralInt(a), LiquidLiteralInt(b):
            return a == b
        case LiquidLiteralFloat(a), LiquidLiteralFloat(b):
            return a == b
        case LiquidLiteralString(a), LiquidLiteralString(b):
            return a == b
        case LiquidApp(a1, a2), LiquidApp(b1, b2):
            return a1 == b1 and all(core_liquid_equality(i2, j2, rename_left) for (i2, j2) in zip(a2, b2))
        case LiquidHornApplication(a1, _), LiquidHornApplication(b1, _):
            return a1 == b1
        case _:
            return False


def core_type_equality(type1: Type, type2: Type, rename_left: dict[Name, Name] | None = None) -> bool:
    "Equality of types up to alpha renaming"
    rename_left = rename_left or {}
    match type1, type2:
        case TypeVar(an), TypeVar(bn):
            return rename_left.get(an, an) == bn
        case RefinedType(aname, aty, aref), RefinedType(bname, bty, bref):
            return core_type_equality(aty, bty, rename_left) and core_liquid_equality(
                aref, bref, rename_left | {aname: bname}
            )
        case AbstractionType(aname, a1, a2), AbstractionType(bname, b1, b2):
            return core_type_equality(a1, b1, rename_left) and core_type_equality(a2, b2, rename_left | {aname: bname})
        case TypeConstructor(aname, a1), TypeConstructor(bname, b1):
            return aname == bname and all(core_type_equality(a, b, rename_left) for a, b in zip(a1, b1))
        case TypePolymorphism(aname, akind, abody), TypePolymorphism(bname, bkind, bbody):
            return akind == bkind and core_type_equality(abody, bbody, rename_left | {aname: bname})
        case Top(), Top():
            return True
        case _:
            return False


def core_term_equality(term1: Term, term2: Term, rename_left: dict[Name, Name] | None = None) -> bool:
    "Equality of terms up to alpha renaming"
    rename_left = rename_left or {}
    match term1, term2:
        case Var(an), Var(bn):
            return rename_left.get(an, an) == bn
        case Literal(av, at), Literal(bv, bt):
            return av == bv and core_type_equality(at, bt, rename_left)
        case Application(a1, a2), Application(b1, b2):
            return core_term_equality(a1, b1, rename_left) and core_term_equality(a2, b2, rename_left)
        case Abstraction(an, abody), Abstraction(bn, bbody):
            return core_term_equality(abody, bbody, rename_left | {an: bn})
        case Let(aname, aval, acont), Let(bname, bval, bcont):
            return core_term_equality(aval, bval, rename_left) and core_term_equality(
                acont, bcont, rename_left | {aname: bname}
            )
        case Rec(aname, atype, aval, acont), Rec(bname, btype, bval, bcont):
            nrename = rename_left | {aname: bname}
            return (
                core_term_equality(aval, bval, nrename)
                and core_type_equality(atype, btype, rename_left)
                and core_term_equality(acont, bcont, nrename)
            )
        case If(ac, at, ao), If(bc, bt, bo):
            return (
                core_term_equality(ac, bc, rename_left)
                and core_term_equality(at, bt, rename_left)
                and core_term_equality(ao, bo, rename_left)
            )
        case TypeApplication(ab, at), TypeApplication(bb, bt):
            return core_term_equality(ab, bb, rename_left) and core_type_equality(at, bt, rename_left)
        case TypeAbstraction(aname, akind, abody), TypeAbstraction(bname, bkind, bbody):
            return akind == bkind and core_term_equality(abody, bbody, rename_left | {aname: bname})
        case _:
            return False



from typing import Any

from sympy import And, Or, Not, Eq, Ne, Lt, Gt, Ge, Le, Symbol
from sympy.core.basic import Basic
from sympy.core.expr import Expr
from sympy.logic.boolalg import to_cnf
from sympy.sets.sets import Set
from sympy.solvers.inequalities import reduce_rational_inequalities

from aeon.core.liquid import (
    LiquidApp,
    LiquidLiteralBool,
    LiquidVar,
    LiquidTerm,
    LiquidLiteralInt,
    LiquidLiteralFloat,
    LiquidLiteralString,
)

sympy_context = {
    "+": lambda x: lambda y: x + y,
    "-": lambda x: lambda y: x - y,
    "*": lambda x: lambda y: x * y,
    "/": lambda x: lambda y: x / y,
    "%": lambda x: lambda y: x % y,
    "%.": lambda x: lambda y: x % y,
    "==": lambda x: lambda y: Eq(x, y),
    "!=": lambda x: lambda y: Ne(x, y),
    "<": lambda x: lambda y: Lt(x, y),
    ">": lambda x: lambda y: Gt(x, y),
    "<=": lambda x: lambda y: Le(x, y),
    ">=": lambda x: lambda y: Ge(x, y),
    # BoolAlg expressions
    "!": lambda x: Not(x),
    "&&": lambda x: lambda y: And(x, y),
    "||": lambda x: lambda y: Or(x, y),
}


def liquid_app_to_sympy(ref: LiquidApp) -> Basic:
    assert ref.fun.id == 0
    ref_fun = ref.fun.name
    assert len(ref.args) == 2
    # if ref_fun in sympy_context:
    #    return sympy_context[ref_fun](*[refined_to_sympy_expression(arg) for arg in ref.args])
    if ref_fun == ">":
        arg0 = refined_to_sympy_expression(ref.args[0])
        arg1 = refined_to_sympy_expression(ref.args[1])
        return Gt(arg0, arg1)
    elif ref_fun == "&&":
        arg0 = refined_to_sympy_expression(ref.args[0])
        arg1 = refined_to_sympy_expression(ref.args[1])
        return And(arg0, arg1)
    elif ref_fun == "||":
        arg0 = refined_to_sympy_expression(ref.args[0])
        arg1 = refined_to_sympy_expression(ref.args[1])
        return Or(arg0, arg1)
    elif ref_fun == "<":
        assert len(ref.args) == 2
        arg0 = refined_to_sympy_expression(ref.args[0])
        arg1 = refined_to_sympy_expression(ref.args[1])
        return Lt(arg0, arg1)
    elif ref_fun == ">=":
        arg0 = refined_to_sympy_expression(ref.args[0])
        arg1 = refined_to_sympy_expression(ref.args[1])
        return Ge(arg0, arg1)
    elif ref_fun == "<=":
        arg0 = refined_to_sympy_expression(ref.args[0])
        arg1 = refined_to_sympy_expression(ref.args[1])
        return Le(arg0, arg1)
    elif ref_fun == "==":
        arg0 = refined_to_sympy_expression(ref.args[0])
        arg1 = refined_to_sympy_expression(ref.args[1])
        return Eq(arg0, arg1)
    elif ref_fun == "!=":
        arg0 = refined_to_sympy_expression(ref.args[0])
        arg1 = refined_to_sympy_expression(ref.args[1])
        return Ne(arg0, arg1)
    else:
        raise ValueError(f"Unknown Liquid function {ref_fun}")


def refined_to_sympy_expression(ref: LiquidTerm) -> Any:
    # ref = ty.refinement
    # name = ty.name
    # base_type_str = ty.type

    match ref:
        case LiquidApp(_, _):
            return liquid_app_to_sympy(ref)
        case LiquidVar(name):
            return Symbol(str(name))
        case LiquidLiteralBool(v) | LiquidLiteralInt(v) | LiquidLiteralFloat(v) | LiquidLiteralString(v):
            return v
        case _:
            raise ValueError(f"Unknown Liquid term {ref} : {type(ref)}")

    # return ty.predicate.to_sympy_expression(ty.variable)


def flatten_conditions(lista: list | Any) -> list:
    if not isinstance(lista, list):
        return [lista]

    return [item for sublist in lista for item in flatten_conditions(sublist)]


def conditional_to_interval(cond: list, name: str) -> Set:
    assert isinstance(name, str)
    try:
        return reduce_rational_inequalities(
            [cond],
            Symbol(name),
            relational=False,
        )
    except Exception as err:
        raise Exception("Failed to do ranged analysis due to: {}".format(err))


def sympy_exp_to_bounded_interval(exp: Expr | Basic) -> Any:
    if isinstance(exp, And):
        return [sympy_exp_to_bounded_interval(x) for x in exp.args]
    elif isinstance(exp, Or):
        return tuple(sympy_exp_to_bounded_interval(x) for x in exp.args)
    elif isinstance(exp, Not):
        # Propagate the not
        exp = to_cnf(exp)
        reduce_rational_inequalities([[exp]], [])
        return [sympy_exp_to_bounded_interval(exp)]
    else:
        return [exp]



from __future__ import annotations

from aeon.core.types import Type
from aeon.sugar.lowering import type_to_core
from aeon.typechecking.context import TypingContext
from aeon.prelude.prelude import typing_vars
from aeon.utils.name import Name


def build_context(ls: dict[Name, Type]) -> TypingContext:
    tc = TypingContext()
    for name, ty in ls.items():
        tc = tc.with_var(name, ty)
    return tc


def built_std_context(ls: dict[Name, Type] | None = None) -> TypingContext:
    data = {k: type_to_core(typing_vars[k]) for k in typing_vars}
    if ls is not None:
        data.update(ls)
    return build_context(data)



"""Meta-programming code for optimization-related decorators."""

from typing import NamedTuple
from aeon.decorators.api import Metadata, metadata_update
from aeon.sugar.program import Definition, STerm, SVar
from aeon.sugar.stypes import STypeConstructor
from aeon.sugar.ast_helpers import st_int, st_float
from aeon.utils.name import Name, fresh_counter

from aeon.sugar.program import SLiteral


def raise_decorator_error(name: str) -> None:
    raise Exception(f"Exception in decorator named {name}.")


class Goal(NamedTuple):
    minimize: bool
    length: int
    function: Name


def make_optimizer(
    args: list[STerm], fun: Definition, metadata: Metadata, typ: STypeConstructor, minimize: bool, length: int = 1
) -> tuple[Definition, list[Definition], Metadata]:
    """This decorator expects a single argument (the body of the definition).

    It does not modify the original definition, but appends a new
    definition to the program. This new definition has the name
    "_fitness_function", prefixed by the original definition's name
    """
    current_goals = metadata.get(fun.name, {}).get("goals", [])
    minimize_name = "minimize" if minimize else "maximize"
    function_name = Name(f"__internal__{minimize_name}_{fun.name}_{len(current_goals)}", fresh_counter.fresh())
    function = Definition(
        name=function_name,
        foralls=[],
        args=[],
        type=typ,
        body=args[0],
    )
    goal = Goal(minimize, length, function_name)

    metadata = metadata_update(
        metadata,
        fun,
        {
            "goals": current_goals + [goal],
        },
    )
    return fun, [function], metadata


def minimize_int(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    assert len(args) == 1, "minimize_int decorator expects a single argument"
    return make_optimizer(args, fun, metadata, st_int, minimize=True)


def maximize_int(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    assert len(args) == 1, "maximize_int decorator expects a single argument"
    return make_optimizer(args, fun, metadata, st_int, minimize=False)


def minimize_float(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    assert len(args) == 1, "minimize_float decorator expects a single argument"
    return make_optimizer(args, fun, metadata, st_float, minimize=True)


def maximize_float(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    assert len(args) == 1, "maximize_float decorator expects a single argument"
    return make_optimizer(args, fun, metadata, st_float, minimize=False)


def multi_minimize_float(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    """This decorator expects a single argument (the body of the definition)."""
    assert (
        len(
            args,
        )
        == 1
    ), "multi_minimize_float decorator expects a single argument"
    assert isinstance(args[1], SLiteral)
    assert isinstance(args[1].value, int), "multi_minimize_float decorator expects an integer argument"
    number_of_objectives = args[1].value
    return make_optimizer(
        args, fun, metadata, STypeConstructor(Name("List", 0)), minimize=True, length=number_of_objectives
    )


def hide(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    """This decorator expects more than zero arguments.

    It does not modify the original definition. It makes sure that no
    grammar nodes are generated from the var names passed as arguments.
    """
    assert len(args) != 0

    # TODO How can I verify if the function is in the context?
    def get_var_name(arg):
        if isinstance(arg, SVar):
            return arg.name
        else:
            raise_decorator_error("hide")

    # rethink this
    aux_dict = {"hide": [get_var_name(arg) for arg in args]}
    metadata = metadata_update(metadata, fun, aux_dict)

    return fun, [], metadata


def hide_types(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    """This decorator expects more than zero arguments.

    It does not modify the original definition. It makes sure that no
    grammar nodes are generated from the var names passed as arguments.
    """
    assert len(args) != 0

    # TODO How can I verify if the function is in the context?
    def get_var_name(arg):
        if isinstance(arg, SVar):
            return arg.name
        else:
            raise_decorator_error("hide_types")

    # rethink this
    aux_dict = {"hide_types": [get_var_name(arg) for arg in args]}
    metadata = metadata_update(metadata, fun, aux_dict)

    return fun, [], metadata


def error_fitness(
    args: list[STerm], fun: Definition, metadata: Metadata
) -> tuple[Definition, list[Definition], Metadata]:
    """This decorator expects one argument .

    It does not modify the original definition. It makes sure that
    the error fitness in case of any exception during the synthesis is the one defined in the argument
    """
    assert len(args) == 1

    aux_dict = {"error_fitness": args[0]}
    metadata = metadata_update(metadata, fun, aux_dict)

    return fun, [], metadata


def disable_control_flow(
    args: list[STerm], fun: Definition, metadata: Metadata
) -> tuple[Definition, list[Definition], Metadata]:
    """This decorator expects zero arguments .

    It does not modify the original definition. It makes sure that
    the control flow grammar nodes are allowed during synthesis
    """
    assert len(args) == 0

    aux_dict = {"disable_control_flow": True}
    metadata = metadata_update(metadata, fun, aux_dict)

    return fun, [], metadata


def allow_recursion(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    """This decorator expects a zero argument .

    It does not modify the original definition. It makes sure that
    recursion can be used during synthesis
    """
    assert len(args) == 0

    aux_dict = {"recursion": True}
    metadata = metadata_update(metadata, fun, aux_dict)

    return fun, [], metadata


def prompt(
    args: list[STerm],
    fun: Definition,
    metadata: Metadata,
) -> tuple[Definition, list[Definition], Metadata]:
    "Keeps track of the prompt that should be used in LLM-based synthesis"
    assert len(args) == 1
    assert isinstance(args[0], SLiteral) and isinstance(args[0].value, str)
    val = args[0].value
    metadata = metadata_update(metadata, fun, {"prompt": val})

    return fun, [], metadata



from __future__ import annotations

from aeon.sugar.stypes import (
    SAbstractionType,
    SRefinedType,
    SType,
    STypeConstructor,
    STypePolymorphism,
    STypeVar,
)
from aeon.sugar.substitutions import normalize
from aeon.utils.name import Name


def type_substitution(ty: SType, alpha: Name, beta: SType) -> SType:
    """t[alpha := beta], standard substition."""

    def rec(x):
        return type_substitution(x, alpha, beta)

    ty = normalize(ty)

    match ty:
        case STypeVar(name):
            if name == alpha:
                return beta
            else:
                return ty
        case SRefinedType(name, ity, ref, loc):
            return normalize(SRefinedType(name, rec(ity), ref, loc=loc))
        case SAbstractionType(name, vty, rty, loc):
            return SAbstractionType(name, rec(vty), rec(rty), loc=loc)
        case STypePolymorphism(name, kind, body, loc):
            # TODO: Double-check alpha_renaming in substitution
            if name == alpha:
                return ty
            else:
                return STypePolymorphism(name, kind, rec(body), loc=loc)
        case STypeConstructor(name, args, loc):
            return STypeConstructor(name, [rec(a) for a in args], loc=loc)
        case _:
            return ty


# TODO: NOW! what is the difference in these two functions?
def type_variable_instantiation(ty: SType, alpha: str, beta: SType) -> SType:
    """t[alpha |-> beta], instantiation."""

    ty = normalize(ty)

    def rec(x: SType):
        return type_variable_instantiation(x, alpha, beta)

    match ty:
        case STypeVar(name):
            if name == alpha:
                return beta
            else:
                return ty
        case SRefinedType(name, ity, ref, loc):
            return normalize(SRefinedType(name, rec(ity), ref, loc=loc))
        case SAbstractionType(var_name, var_type, ret_type, loc):
            return SAbstractionType(var_name, rec(var_type), rec(ret_type), loc=loc)
        case STypePolymorphism(name, kind, body, loc):
            if name == alpha:
                return ty
            else:
                return STypePolymorphism(name, kind, rec(body), loc=loc)
        case STypeConstructor(name, args, loc):
            return STypeConstructor(name, [rec(a) for a in args], loc=loc)
        case _:
            assert False



from abc import ABC
from dataclasses import dataclass, field
from pathlib import Path
from typing import Iterable

from aeon.core.terms import Term
from aeon.core.types import Kind, Type
from aeon.sugar.program import ImportAe, STerm
from aeon.sugar.stypes import SType
from aeon.typechecking.context import TypingContext
from aeon.utils.location import Location
from aeon.verification.vcs import Constraint
from aeon.verification.helpers import pretty_print_constraint


class AeonError(ABC, BaseException):
    def __str__(self) -> str: ...

    def position(self) -> Location: ...


@dataclass
class ImportError(AeonError):
    importel: ImportAe
    possible_containers: Iterable[Path]

    def __str__(self) -> str:
        return f"Could not find {self.importel.path} in the following list of container folders: {self.possible_containers}"

    def position(self) -> Location:
        return self.importel.loc


@dataclass
class UnificationSubtypingError(AeonError):
    expr: STerm
    subtype: SType
    suptype: SType
    msg: str = field(default_factory=lambda: "")

    def __str__(self) -> str:
        return (
            f"Expression {self.expr} has type {self.subtype}, but is expected to have type {self.suptype} ({self.msg})."
        )

    def position(self) -> Location:
        return self.expr.loc


@dataclass
class UnificationFailedError(AeonError):
    name: str
    conflict1: SType
    conflict2: SType

    def __str__(self) -> str:
        return f"Unification variable {self.name} needs to be {self.conflict1} and {self.conflict2} simultaneously, but the two types are not compatible."

    def position(self) -> Location:
        return self.conflict1.loc


@dataclass
class UnificationKindError(AeonError):
    term: STerm
    type: SType
    kind1: Kind
    kind2: Kind

    def __str__(self) -> str:
        return f"Term {self.term} has type {self.type} with kind {self.kind1}, but kind {self.kind2} is expected."

    def position(self) -> Location:
        return self.term.loc


@dataclass
class UnificationUnknownTypeError(AeonError):
    term: STerm

    def __str__(self) -> str:
        return f"Variable {self.term} does not exist in context."

    def position(self) -> Location:
        return self.term.loc


class CoreTypeCheckingError(AeonError):
    pass


@dataclass
class CoreWellformnessError(CoreTypeCheckingError):
    type: Type

    def __str__(self) -> str:
        return f"Type {self.type} is not well formed."

    def position(self) -> Location:
        return self.type.loc


@dataclass
class LiquidTypeCheckingFailedRelation(CoreTypeCheckingError):
    ctx: TypingContext
    term: Term
    type: Type
    vc: Constraint

    def __str__(self) -> str:
        return f"Failed to prove ({pretty_print_constraint(self.vc)}) in {self.position()}"

    def position(self) -> Location:
        return self.term.loc


@dataclass
class CoreTypingRelation(CoreTypeCheckingError):
    ctx: TypingContext
    term: Term
    type: Type

    def __str__(self) -> str:
        return f"Term {self.term} is not of expected type {self.type}."

    def position(self) -> Location:
        return self.term.loc


@dataclass
class CoreVariableNotInContext(CoreTypeCheckingError):
    ctx: TypingContext
    term: Term

    def __str__(self) -> str:
        return f"Variable {self.term} is not in context."

    def position(self) -> Location:
        return self.term.loc


@dataclass
class CoreInvalidApplicationError(CoreTypeCheckingError):
    term: Term
    type: Type

    def __str__(self) -> str:
        return f"Function application requires that {self.term} is a function, but it has type {self.type}."

    def position(self) -> Location:
        return self.term.loc


@dataclass
class CoreTypeApplicationRequiresBareTypesError(CoreTypeCheckingError):
    term: Term
    type: Type

    def __str__(self) -> str:
        return f"Type application requires that {self.term} of type {self.type} has a bare kind."

    def position(self) -> Location:
        return self.term.loc


@dataclass
class CoreWrongKindInTypeApplicationError(CoreTypeCheckingError):
    term: Term
    type: Type
    expected_kind: Kind
    actual_kind: Kind

    def __str__(self) -> str:
        return f"Type application requires that {self.term} of type {self.type} has kind {self.expected_kind}, but {self.actual_kind} was found."

    def position(self) -> Location:
        return self.term.loc


@dataclass
class CoreSubtypingError(CoreTypeCheckingError):
    ctx: TypingContext
    term: Term
    inferred_type: Type
    expected_type: Type

    def __str__(self) -> str:
        return f"Expression {self.term} is expected to be of type {self.expected_type}, but {self.inferred_type} was found instead."

    def position(self) -> Location:
        return self.term.loc



from __future__ import annotations

import argparse
import sys
from argparse import ArgumentParser

from aeon.facade.api import AeonError
from aeon.facade.driver import AeonConfig, AeonDriver
from aeon.logger.logger import export_log
from aeon.logger.logger import setup_logger
from aeon.lsp.server import AeonLanguageServer
from aeon.synthesis.uis.api import SynthesisUI, SynthesisFormat
from aeon.synthesis.uis.terminal import TerminalUI
from aeon.utils.pprint import pretty_print_sterm

sys.setrecursionlimit(10000)


def parse_arguments():
    parser = argparse.ArgumentParser()

    if "-lsp" in sys.argv or "--language-server-mode" in sys.argv:
        parser.add_argument(
            "-lsp",
            "--language-server-mode",
            action="store_true",
            help="run language server mode",
        )
        parser.add_argument(
            "--tcp",
            help="listen on tcp port or hostname:port on IPv4.",
            type=str,
        )

    else:
        parser.add_argument("filename", help="name of the aeon files to be synthesized")

    _parse_common_arguments(parser)
    return parser.parse_args()


def _parse_common_arguments(parser: ArgumentParser):
    parser.add_argument("--core", action="store_true", help="synthesize a aeon core file")
    parser.add_argument("--budget", type=int, default=60, help="Time for synthesis (in seconds).")
    parser.add_argument(
        "-l",
        "--log",
        nargs="+",
        default="",
        help="""set log level: \nTRACE \nDEBUG \nINFO \nWARNINGS \nCONSTRAINT \nTYPECHECKER \nSYNTH_TYPE \nCONSTRAINT \nSYNTHESIZER
                        \nERROR \nCRITICAL\n TIME""",
    )
    parser.add_argument(
        "-f",
        "--logfile",
        action="store_true",
        help="export log file",
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Show debug information",
    )

    parser.add_argument(
        "-t",
        "--timings",
        action="store_true",
        help="Show timing information",
    )

    parser.add_argument("-n", "--no-main", action="store_true", help="Disables introducing hole in main")

    parser.add_argument(
        "-s",
        "--synthesizer",
        type=str,
        default="gp",
        help="Select a synthesizer for synthesis(gp for Genetic programming(Defaut), synquid for Synquid, random_search for Random Search, enumerative for Enumerative Search, hc for Hill Climbing, and 1p1 for One Plus One)",
    )

    parser.add_argument(
        "--synthesis-format",
        type=str,
        choices=["default", "json"],
        default="default",
        help="Select the synthesised holes format results: default or json",
    )

    parser.add_argument(
        "--format",
        action="store_true",
        help="Prints a pretty print version of the code to the stdout",
    )

    parser.add_argument(
        "--fix",
        action="store_true",
        help="Uses a pretty print version of the code to reformat it",
    )

    return parser.parse_args()


def select_synthesis_ui() -> SynthesisUI:
    return TerminalUI()


def handle_error(err: AeonError):
    # TODO: handle each error with proper printing
    match err:
        case _:
            print(f">>> Error at {err.position()}:")
            print(err)


def main() -> None:
    args = parse_arguments()

    logger = setup_logger()
    logfile_name = None
    if hasattr(args, "filename"):
        logfile_name = args.filename
    elif hasattr(args, "language_server_mode"):
        logfile_name = "lsp"
    export_log(args.log, args.logfile, logfile_name)

    if args.debug:
        logger.add(sys.stderr)
    if args.timings:
        logger.add(sys.stderr, level="TIME")

    cfg = AeonConfig(
        synthesizer=args.synthesizer,
        synthesis_ui=select_synthesis_ui(),
        synthesis_budget=args.budget,
        timings=args.timings,
        no_main=args.no_main,
        synthesis_format=SynthesisFormat.from_string(args.format),
    )
    driver = AeonDriver(cfg)

    if hasattr(args, "language_server_mode"):
        aeon_lsp = AeonLanguageServer(driver)
        aeon_lsp.start(args.tcp)
        sys.exit(0)

    if args.core:
        errors = driver.parse_core(args.filename)
    else:
        errors = driver.parse(args.filename)

    if errors:
        for err in errors:
            handle_error(err)
    elif args.format:
        driver.pretty_print(args.filename, args.fix)

    elif driver.has_synth():
        term = driver.synth()
        print("Synthesized:")
        print("#str")
        print(str(term))
        print("#pprint")
        print(pretty_print_sterm(term))
    else:
        driver.run()


if __name__ == "__main__":
    main()



from aeon.core.liquid import LiquidApp, LiquidTerm, LiquidVar
from aeon.core.substitutions import substitution_in_liquid, substitution_in_type
from aeon.core.terms import Var
from aeon.core.types import AbstractionType, Type
from aeon.core.types import TypeConstructor
from aeon.core.types import RefinedType
from aeon.core.types import Top
from aeon.core.pprint import aeon_prelude_ops_to_text
from aeon.utils.name import Name


def mangle_name(name: Name) -> str:
    if name.id == 0:
        return name.name
    elif name.id == -1:
        return name.name + "_unknown"
    return name.name + "_" + str(name.id)


def mangle_liquid_term(refinement: LiquidTerm) -> str:
    match refinement:
        case LiquidApp(Name(fun, _), args):
            fun_name = aeon_prelude_ops_to_text.get(fun, fun)
            args_text = "__comma__".join([mangle_liquid_term(t) for t in args])
            return f"__opar__{fun_name}__{args_text}__cpar__"
        case _:
            return str(refinement)


def mangle_var(name: Name) -> str:
    """Mangles the name of variable, so it is a valid Python identifier."""
    return aeon_prelude_ops_to_text.get(name.name, mangle_name(name))


def mangle_type(ty: Type) -> str:
    """Mangles the Aeon Type name, to be a valid Python name."""
    match ty:
        case TypeConstructor(name, _):
            return f"æ{mangle_name(name)}"
        case RefinedType(name, ty, ref):
            ref2 = substitution_in_liquid(ref, LiquidVar(Name("__self__", 0)), name)
            return f"_{mangle_type(ty)}_{mangle_liquid_term(ref2)}"
        case AbstractionType(var_name, var_type, type):
            ty1 = mangle_type(var_type)
            ty2 = mangle_type(substitution_in_type(type, Var(Name("__self__", 0)), var_name))
            return f"{ty1}_arrow_{ty2}"
        case Top():
            return "Top"
        case _:
            assert False, f"Unsupported {ty}"



from abc import ABC
from dataclasses import dataclass, field
from typing import MutableSequence

from aeon.core.types import Kind
from aeon.sugar.program import TypeDecl
from aeon.sugar.stypes import SType
from aeon.utils.name import Name, fresh_counter


class ElabTypingContextEntry(ABC):
    pass


@dataclass
class ElabVariableBinder(ElabTypingContextEntry):
    name: Name
    type: SType


@dataclass
class ElabUninterpretedBinder(ElabTypingContextEntry):
    name: Name
    type: SType


@dataclass
class ElabTypeVarBinder(ElabTypingContextEntry):
    name: Name
    type: Kind


@dataclass
class ElabTypeDecl(ElabTypingContextEntry):
    name: Name
    args: list[Name]


@dataclass
class ElaborationTypingContext:
    entries: MutableSequence[ElabTypingContextEntry] = field(default_factory=list)

    def type_of(self, name: Name):
        """Returns the type of the variable name."""
        for entry in self.entries[::-1]:
            match entry:
                case ElabVariableBinder(bname, ty):
                    if bname.name == name.name:
                        return ty
        return None

    def with_var(self, name: Name, ty: SType):
        """Creates a new context, with an extra variable."""
        nentries = [x for x in self.entries] + [ElabVariableBinder(name, ty)]
        return ElaborationTypingContext(nentries)

    def with_typevar(self, name: Name, kind: Kind):
        """Creates a new context, with an extra type variable"""
        nentries = [x for x in self.entries] + [ElabTypeVarBinder(name, kind)]
        return ElaborationTypingContext(nentries)

    def fresh_typevar(self) -> Name:
        """Returns a type variable that does not exist in context."""
        return Name("tv", fresh_counter.fresh())


def build_typing_context(ls: dict[Name, SType], tdecl: list[TypeDecl] | None = None) -> ElaborationTypingContext:
    if tdecl is None:
        tdecl = []
    return ElaborationTypingContext(
        [ElabVariableBinder(name, ls[name]) for name in ls] + [ElabTypeDecl(td.name, td.args) for td in tdecl]
    )



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, Iterable

DEFAULT_NEW_LINE_CHAR = "\n"
DEFAULT_SPACE_CHAR = " "
DEFAULT_WIDTH = 120
DEFAULT_TAB_SIZE = 4


class Doc(ABC):
    @abstractmethod
    def layout(self, indent: int) -> str: ...

    @abstractmethod
    def fits(self, width: int, current_length: int) -> bool: ...

    @abstractmethod
    def best(self, width: int, current_length: int) -> "Doc": ...

    @abstractmethod
    def flatten(self) -> "Doc": ...

    def __str__(self):
        return self.best(DEFAULT_WIDTH, 0).layout(0)

    def __add__(self, other):
        return Concat(self, other)

    def calculate_new_length(self, current_length: int) -> int:
        match self:
            case Line():
                return 0
            case Text(value):
                return current_length + len(value)
            case _:
                layout = self.layout(0)
                if DEFAULT_NEW_LINE_CHAR in layout:
                    last_line = layout.rsplit(DEFAULT_NEW_LINE_CHAR, 1)[-1]
                    return len(last_line)
                else:
                    return current_length + len(layout)


@dataclass(frozen=True)
class Nil(Doc):
    def layout(self, _) -> str:
        return ""

    def fits(self, *_) -> bool:
        return True

    def best(self, *_) -> "Doc":
        return self

    def flatten(self) -> "Doc":
        return self


@dataclass(frozen=True)
class Text(Doc):
    value: str

    def layout(self, _) -> str:
        return self.value

    def fits(self, width: int, current_length: int) -> bool:
        return current_length + len(self.value) <= width

    def best(self, *_) -> "Doc":
        return self

    def flatten(self) -> "Doc":
        return self


@dataclass(frozen=True)
class Line(Doc):
    def layout(self, indent: int) -> str:
        return DEFAULT_NEW_LINE_CHAR + DEFAULT_SPACE_CHAR * indent

    def fits(self, width: int, current_length: int) -> bool:
        return current_length + 1 <= width

    def best(self, *_) -> "Doc":
        return self

    def flatten(self) -> "Doc":
        return Text(DEFAULT_SPACE_CHAR)


@dataclass(frozen=True)
class SoftLine(Line):
    def fits(self, width: int, current_length: int) -> bool:
        return current_length <= width

    def flatten(self) -> "Doc":
        return Text("")


@dataclass(frozen=True)
class HardLine(Line):
    def fits(self, *_) -> bool:
        return False

    def flatten(self) -> "Doc":
        return self


@dataclass(frozen=True)
class Concat(Doc):
    left: Doc
    right: Doc

    def layout(self, indent: int) -> str:
        return self.left.layout(indent) + self.right.layout(indent)

    def fits(self, width: int, current_length: int) -> bool:
        if not self.left.fits(width, current_length):
            return False

        new_length = self.left.calculate_new_length(current_length)

        return self.right.fits(width, new_length)

    def best(self, width: int, current_length: int) -> "Doc":
        left_best = self.left.best(width, current_length)

        new_length = left_best.calculate_new_length(current_length)

        right_best = self.right.best(width, new_length)
        return Concat(left_best, right_best)

    def flatten(self) -> "Doc":
        return self.left.flatten() + self.right.flatten()


@dataclass(frozen=True)
class MultiUnion(Doc):
    alternatives_fn: Callable[[], Iterable[Doc]]

    def layout(self, indent: int) -> str:
        return self.best(DEFAULT_WIDTH, 0).layout(indent)

    def fits(self, width: int, current_length: int) -> bool:
        return any(doc.fits(width, current_length) for doc in self.alternatives_fn())

    def best(self, width: int, current_length: int) -> "Doc":
        default_document = None
        for doc in self.alternatives_fn():
            if doc.fits(width, current_length):
                return doc.best(width, current_length)
            default_document = doc
        return default_document.best(width, current_length)

    def flatten(self) -> "Doc":
        for doc in self.alternatives_fn():
            return doc.flatten()
        return nil()


@dataclass(frozen=True)
class Nest(Doc):
    indent: int
    doc: Doc

    def layout(self, indent: int) -> str:
        return self.doc.layout(indent + self.indent)

    def fits(self, width: int, current_length: int) -> bool:
        return self.doc.fits(width, current_length)

    def best(self, width: int, current_length: int) -> "Doc":
        return Nest(self.indent, self.doc.best(width, current_length + self.indent))

    def flatten(self) -> "Doc":
        return Nest(self.indent, self.doc.flatten())


def nil() -> Doc:
    return Nil()


def text(value: str) -> Doc:
    return Text(value) if value != "" else nil()


def line() -> Doc:
    return Line()


def soft_line() -> Doc:
    return SoftLine()


def hard_line() -> Doc:
    return HardLine()


def concat(docs: list[Doc]) -> Doc:
    filtered_docs = [doc for doc in docs if not isinstance(doc, Nil)]

    if not filtered_docs:
        return nil()

    result = filtered_docs[0]
    for doc in filtered_docs[1:]:
        result += doc

    return result


def nest(indent: int, doc: Doc) -> Doc:
    return Nest(indent, doc)


def group(doc: Doc) -> Doc:
    return MultiUnion(lambda: iter([doc.flatten(), doc]))


def parens(doc: Doc, needs_spaces: bool = False) -> Doc:
    line_func = line if needs_spaces else soft_line
    return group(concat([text("("), nest(DEFAULT_TAB_SIZE, concat([line_func(), doc])), line_func(), text(")")]))


def insert_between(separator: Doc, docs: list[Doc]) -> Doc:
    if not docs:
        return nil()

    docs_with_separator = docs[0]
    for doc in docs[1:]:
        docs_with_separator = concat([docs_with_separator, separator, doc])

    return docs_with_separator



from __future__ import annotations

import zss

from aeon.core.terms import Abstraction
from aeon.core.terms import Application
from aeon.core.terms import If
from aeon.core.terms import Let
from aeon.core.terms import Literal
from aeon.core.terms import Rec
from aeon.core.terms import Term
from aeon.core.terms import Var


def term_children(t: Term):
    if isinstance(t, Application):
        return [t.fun, t.arg]
    elif isinstance(t, Abstraction):
        return [t.var_name, t.body]
    elif isinstance(t, Rec):
        return [t.var_name, t.var_value, t.body]  # Type?
    elif isinstance(t, Let):
        return [t.var_name, t.var_value, t.body]
    elif isinstance(t, If):
        return [t.cond, t.then, t.otherwise]
    return []


def term_label(t: Term):
    if isinstance(t, Literal):
        return str(t.value)
    elif isinstance(t, Var):
        return str(t.name)
    return type(t).__name__


def term_label_dist(s1: str, s2: str):
    if s1 == s2:
        return 0
    else:
        return 1


def distance_terms(t1: Term, t2: Term):
    return zss.simple_distance(t1, t2, term_children, term_label, term_label_dist)


def pairwise_distance(ts: list[Term]):
    s = 0
    for i in range(len(ts)):
        for j in range(i + 1, len(ts)):
            t1 = ts[i]
            t2 = ts[j]
            s += distance_terms(t1, t2)
    return s


def term_depth(t: Term):
    if isinstance(t, Application):
        return 1 + max(term_depth(t.fun), term_depth(t.arg))
    elif isinstance(t, Abstraction):
        return 1 + term_depth(t.body)
    elif isinstance(t, Rec):
        return 1 + max(term_depth(t.var_value), term_depth(t.body))
    elif isinstance(t, Let):
        return 1 + max(term_depth(t.var_value), term_depth(t.body))
    elif isinstance(t, If):
        return 1 + max(term_depth(t.cond), term_depth(t.then), term_depth(t.otherwise))
    else:
        return 1



from time import monotonic_ns
from typing import Any, Callable

from aeon.core.terms import Term
from aeon.core.types import Type
from aeon.decorators.api import Metadata
from aeon.synthesis.api import Synthesizer
from aeon.synthesis.modules.synquid.build import synthes_memory
from aeon.synthesis.uis.api import SynthesisUI
from aeon.typechecking.context import TypingContext
from aeon.utils.name import Name


def is_better(v1, v2):
    if not v2:
        return True
    return all(x < y for x, y in zip(v1, v2))


def get_elapsed_time(start_time) -> float:
    """The elapsed time since the start in seconds."""
    return (monotonic_ns() - start_time) * 0.000000001


class SynquidSynthesizer(Synthesizer):
    def synthesize(
        self,
        ctx: TypingContext,
        type: Type,
        validate: Callable[[Term], bool],
        evaluate: Callable[[Term], list[float]],
        fun_name: Name,
        metadata: Metadata,
        budget: float = 60,
        ui: SynthesisUI = SynthesisUI(),
    ) -> Term:
        assert isinstance(ctx, TypingContext)
        assert isinstance(type, Type)

        current_metadata = metadata.get(fun_name, {})
        is_recursion_allowed = current_metadata.get("recursion", False)
        vars_to_ignore = current_metadata.get("hide", [])

        def skip(name: Name) -> bool:
            if name == fun_name:
                return not is_recursion_allowed
            elif name in vars_to_ignore:
                return True
            elif name.name.startswith("__internal__"):
                return True
            elif name.name in ["native", "native_import", "print"]:
                return True
            else:
                return False

        start_time = monotonic_ns()
        done = True
        level = 0
        best: tuple[list[float], Any] = ([], None)
        mem: dict = {}
        ui.register(None, None, 0, True)
        while done:
            for result in synthes_memory(ctx, level, type, skip, mem):
                if validate(result):
                    score = evaluate(result)
                    if is_better(score, best[0]):
                        best = (score, result)
                        ui.register(result, score, get_elapsed_time(start_time), True)
                    else:
                        ui.register(result, score, get_elapsed_time(start_time), False)
                else:
                    ui.register(result, "Invalid", get_elapsed_time(start_time), False)
                if get_elapsed_time(start_time) > budget:
                    done = False
                    break
            level += 1
        return best[1]



from typing import Type as TypingType

from geneticengine.grammar.metahandlers.base import MetaHandlerGenerator
from geneticengine.grammar.metahandlers.floats import FloatRange
from geneticengine.grammar.metahandlers.ints import IntRange
from geneticengine.grammar.metahandlers.strings import StringSizeBetween

from aeon.core.liquid import LiquidLiteralInt, LiquidLiteralString, LiquidLiteralFloat
from aeon.core.types import t_bool
from aeon.core.types import t_float
from aeon.core.types import t_int
from aeon.core.types import t_string
from aeon.core.types import Type
from aeon.core.pprint import aeon_prelude_ops_to_text

prelude_ops: list[str] = ["print", "native_import", "native"]

internal_functions: list[str] = []

text_to_aeon_prelude_ops: dict[str, str] = {v: k for k, v in aeon_prelude_ops_to_text.items()}

grammar_base_types: list[str] = ["t_Float", "t_Int", "t_String", "t_Bool"]

aeon_to_python_types: dict[str, TypingType] = {"Int": int, "Bool": bool, "String": str, "Float": float}

aeon_to_python: dict[Type, TypingType] = {
    t_bool: bool,
    t_int: int,
    t_string: str,
    t_float: float,
}

aeon_to_gengy_metahandlers: dict[Type, MetaHandlerGenerator] = {
    t_int: IntRange,
    t_string: StringSizeBetween,
    t_float: FloatRange,
}

aeon_to_liquid_terms: dict[str, TypingType[LiquidLiteralFloat | LiquidLiteralInt | LiquidLiteralString]] = {
    "Int": LiquidLiteralInt,
    "String": LiquidLiteralString,
    "Float": LiquidLiteralFloat,
}



from __future__ import annotations

import math

import numpy as np
from PIL import Image
from PIL.ImageDraw import Draw
from skimage.metrics import mean_squared_error as mse

from aeon.bindings.binding_utils import curried


@curried
def Image_mk(w, h, c):
    return Image.new("RGB", (w, h), c)


@curried
def Image_draw_rectangle(im, x, y, w, h, c):
    im2 = im.copy()
    d = Draw(im2)
    d.rectangle((x, y, x + w, y + h), c)
    return im2


@curried
def Image_diff(im1, im2):
    s = sum((a - b) ** 2 for a, b in zip(im1.histogram(), im2.histogram()))
    return math.sqrt(s)


@curried
def Image_diff_mse(im1, im2):
    im1 = np.array(im1)
    im2 = np.array(im2)
    return mse(im1, im2)



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass, field

from aeon.core.liquid import LiquidLiteralFloat, LiquidLiteralInt, LiquidLiteralString, liquid_free_vars
from aeon.core.liquid import LiquidHole
from aeon.core.liquid import LiquidLiteralBool
from aeon.core.liquid import LiquidTerm
from aeon.utils.location import Location, SynthesizedLocation
from aeon.utils.name import fresh_counter, Name


# TODO: convert to ENUM
class Kind(ABC):
    def __repr__(self):
        return str(self)


class BaseKind(Kind):
    def __eq__(self, o):
        return self.__class__ == o.__class__

    def __str__(self):
        return "Β"

    def __hash__(self):
        return super().__hash__()


class StarKind(Kind):
    def __eq__(self, o):
        return self.__class__ == o.__class__

    def __str__(self):
        return "★"

    def __hash__(self):
        return super().__hash__()


class Type(ABC):
    loc: Location


@dataclass
class TypeVar(Type):
    name: Name
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __post_init__(self):
        if self.name.name in ["Int", "Bool"]:
            assert False

    def __repr__(self):
        return f"{self.name}"

    def __eq__(self, other):
        return isinstance(other, TypeVar) and other.name == self.name

    def __hash__(self) -> int:
        return hash(self.name)


class Top(Type):
    def __repr__(self):
        return "⊤"

    def __eq__(self, other):
        return isinstance(other, Top)

    def __str__(self):
        return repr(self)

    def __hash__(self) -> int:
        return hash("Top")


@dataclass
class AbstractionType(Type):
    var_name: Name
    var_type: Type
    type: Type
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __repr__(self):
        return f"({self.var_name}:{self.var_type}) -> {self.type}"

    def __eq__(self, other):
        return (
            isinstance(other, AbstractionType)
            and self.var_name == other.var_name
            and self.var_type == other.var_type
            and self.type == other.type
        )

    def __hash__(self) -> int:
        return hash(self.var_name) + hash(self.var_type) + hash(self.type)


@dataclass
class RefinedType(Type):
    name: Name
    type: TypeConstructor | TypeVar
    refinement: LiquidTerm
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __repr__(self):
        return f"{{ {self.name}:{self.type} | {self.refinement} }}"

    def __eq__(self, other):
        return (
            isinstance(other, RefinedType)
            and self.name == other.name
            and self.type == other.type
            and self.refinement == other.refinement
        )

    def __hash__(self) -> int:
        return hash(self.name) + hash(self.type) + hash(self.refinement)


@dataclass
class TypePolymorphism(Type):
    name: Name  # alpha
    kind: Kind
    body: Type
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"forall {self.name}:{self.kind}, {self.body}"

    def __hash__(self) -> int:
        return hash(self.name) + hash(self.kind) + hash(self.body)


@dataclass
class TypeConstructor(Type):
    name: Name
    args: list[Type] = field(default_factory=list)
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        args = ", ".join(str(a) for a in self.args)
        return f"{self.name} {args}"

    def __eq__(self, other):
        return isinstance(other, TypeConstructor) and other.name == self.name and self.args == other.args

    def __hash__(self) -> int:
        return hash(self.name) + sum(hash(a) for a in self.args)


# Default type constructors


t_unit = TypeConstructor(Name("Unit", 0), [])
t_bool = TypeConstructor(Name("Bool", 0), [])
t_int = TypeConstructor(Name("Int", 0), [])
t_float = TypeConstructor(Name("Float", 0), [])
t_string = TypeConstructor(Name("String", 0), [])

builtin_core_types = [t_unit, t_bool, t_int, t_float, t_string]

top = Top()


# This class is here to prevent circular imports.


@dataclass
class LiquidHornApplication(LiquidTerm):
    name: Name
    argtypes: list[tuple[LiquidTerm, TypeConstructor | TypeVar]]
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __post_init__(self):
        assert isinstance(self.name, Name)
        for term, ty in self.argtypes:
            match term:
                case LiquidLiteralBool(_):
                    assert ty == TypeConstructor(Name("Bool", 0))
                case LiquidLiteralInt(_):
                    assert ty == TypeConstructor(Name("Int", 0))
                case LiquidLiteralFloat(_):
                    assert ty == TypeConstructor(Name("Float", 0))
                case LiquidLiteralString(_):
                    assert ty == TypeConstructor(Name("String", 0))

    def __repr__(self):
        j = ", ".join([f"{n}:{t}" for (n, t) in self.argtypes])
        return f"?{self.name}({j})"

    def __eq__(self, other):
        return isinstance(other, LiquidHornApplication) and other.name == self.name

    def __hash__(self) -> int:
        return hash(self.name)


liq_true = LiquidLiteralBool(True)


def extract_parts(t: Type) -> tuple[Name, TypeConstructor | TypeVar, LiquidTerm]:
    assert (
        isinstance(t, TypeConstructor)
        or isinstance(t, RefinedType)
        or isinstance(
            t,
            TypeVar,
        )
        or isinstance(t, TypeConstructor)
    )
    match t:
        case RefinedType(name, ity, ref):
            return (name, ity, ref)
        case _:
            return (Name("_", fresh_counter.fresh()), t, liq_true)


def is_bare(t: Type) -> bool:
    """Returns whether the type is bare."""
    match t:
        case TypeConstructor(_, _) | Top() | TypeVar():
            return True
        case RefinedType(_, _, ref):
            return ref == LiquidHole() or isinstance(ref, LiquidHornApplication)
        case AbstractionType(_, _, vtype):
            return is_bare(vtype) and is_bare(vtype)
        case TypePolymorphism(_, _, ty):
            return is_bare(ty)
        case _:
            assert False, f"Unknown type {t} ({type(t)})"


def base(ty: Type) -> Type:
    """Returns the base type of a Refined Type"""
    if isinstance(ty, RefinedType):
        return ty.type
    return ty


def type_free_term_vars(t: Type) -> list[Name]:
    from aeon.prelude.prelude import ALL_OPS

    if isinstance(t, TypeConstructor):
        return []
    elif isinstance(t, TypeVar):
        return []
    elif isinstance(t, AbstractionType):
        afv = type_free_term_vars(t.var_type)
        rfv = type_free_term_vars(t.type)
        return [x for x in afv + rfv if x != t.var_name and x not in ALL_OPS]
    elif isinstance(t, RefinedType):
        ifv = type_free_term_vars(t.type)
        rfv = liquid_free_vars(t.refinement)
        return [x for x in ifv + rfv if x != t.name]
    elif isinstance(t, TypePolymorphism):
        return type_free_term_vars(t.body)
    return []


def get_type_vars(t: Type) -> set[TypeVar]:
    if isinstance(t, TypeConstructor):
        return set()
    elif isinstance(t, TypeVar):
        return {t}
    elif isinstance(t, AbstractionType):
        return get_type_vars(t.var_type).union(get_type_vars(t.type))
    elif isinstance(t, RefinedType):
        return get_type_vars(t.type)
    elif isinstance(t, TypePolymorphism):
        return {t1 for t1 in get_type_vars(t.body) if t1.name != t.name}
    else:
        assert False, f"Unable to extract {t} ({type(t)})"


def refined_to_unrefined_type(ty: Type) -> Type:
    if isinstance(ty, RefinedType):
        return ty.type
    if isinstance(ty, AbstractionType):
        return AbstractionType(
            ty.var_name,
            refined_to_unrefined_type(ty.var_type),
            refined_to_unrefined_type(ty.type),
        )
    return ty



import sys
from dataclasses import dataclass
from functools import reduce
from typing import Any, Iterable

from aeon.backend.evaluator import EvaluationContext
from aeon.backend.evaluator import eval
from aeon.core.bind import bind_ids
from aeon.core.substitutions import substitution
from aeon.core.terms import Term
from aeon.core.types import top
from aeon.decorators import Metadata
from aeon.elaboration import elaborate
from aeon.facade.api import AeonError
from aeon.frontend.anf_converter import ensure_anf
from aeon.frontend.parser import parse_term
from aeon.prelude.prelude import evaluation_vars
from aeon.prelude.prelude import typing_vars
from aeon.sugar.ast_helpers import st_top
from aeon.sugar.bind import bind, bind_program
from aeon.sugar.desugar import DesugaredProgram, desugar
from aeon.sugar.lifting import lift
from aeon.sugar.lowering import lower_to_core, lower_to_core_context, type_to_core
from aeon.sugar.parser import parse_main_program
from aeon.sugar.program import Program, STerm
from aeon.synthesis.entrypoint import synthesize_holes
from aeon.synthesis.identification import incomplete_functions_and_holes
from aeon.synthesis.modules.synthesizerfactory import make_synthesizer
from aeon.synthesis.uis.api import SynthesisUI, SynthesisFormat
from aeon.typechecking.typeinfer import check_type_errors
from aeon.utils.ctx_helpers import build_context
from aeon.utils.name import Name
from aeon.utils.pprint import pretty_print_node
from aeon.utils.time_utils import RecordTime


def read_file(filename: str) -> str:
    with open(filename) as file:
        return file.read()


@dataclass
class AeonConfig:
    synthesizer: str
    synthesis_ui: SynthesisUI
    synthesis_budget: int
    timings: bool = False
    no_main: bool = False
    synthesis_format: SynthesisFormat = SynthesisFormat.DEFAULT


class AeonDriver:
    def __init__(self, cfg: AeonConfig):
        self.cfg = cfg

    def parse_core(self, filename: str):
        # TODO: deprecate core parsing
        aeon_code = read_file(filename)

        core_typing_vars: dict[Name, Any] = reduce(
            lambda acc, el: acc | {el[0]: type_to_core(el[1], available_vars=[e for e in acc.items()])},
            typing_vars.items(),
            {},
        )

        self.typing_ctx = build_context(core_typing_vars)
        self.core_ast = parse_term(aeon_code)
        self.metadata: Metadata = {}

    def parse(self, filename: str = None, aeon_code: str = None) -> Iterable[AeonError]:
        if aeon_code is None:
            aeon_code = read_file(filename)

        with RecordTime("ParseSugar"):
            prog: Program = parse_main_program(aeon_code, filename=filename)
            prog = bind_program(prog, [])

        with RecordTime("Desugar"):
            try:
                desugared: DesugaredProgram = desugar(prog, is_main_hole=not self.cfg.no_main)
            except AeonError as e:
                return [e]

        with RecordTime("Bind"):
            ctx, progt = bind(desugared.elabcontext, desugared.program)
            desugared = DesugaredProgram(progt, ctx, desugared.metadata)
            metadata: Metadata = desugared.metadata

        try:
            with RecordTime("Elaboration"):
                sterm: STerm = elaborate(desugared.elabcontext, desugared.program, st_top)
        except AeonError as e:
            return [e]  # TODO: Support multiple errors

        with RecordTime("Core generation"):
            typing_ctx = lower_to_core_context(desugared.elabcontext)
            core_ast = lower_to_core(sterm)
            typing_ctx, core_ast = bind_ids(typing_ctx, core_ast)

        with RecordTime("ANF conversion"):
            core_ast_anf = ensure_anf(core_ast)

        with RecordTime("TypeChecking"):
            type_errors = check_type_errors(typing_ctx, core_ast_anf, top)
            # TODO
            if type_errors:
                return type_errors

        with RecordTime("Preparing execution env"):
            evaluation_ctx = EvaluationContext(evaluation_vars)

        self.metadata = metadata
        self.core = core_ast_anf
        self.typing_ctx = typing_ctx
        self.evaluation_ctx = evaluation_ctx
        return []

    def run(self) -> None:
        with RecordTime("Evaluation"):
            eval(self.core, self.evaluation_ctx)

    def has_synth(self) -> bool:
        with RecordTime("DetectSynthesis"):
            self.incomplete_functions: list[
                tuple[
                    Name,
                    list[Name],
                ]
            ] = incomplete_functions_and_holes(
                self.typing_ctx,
                self.core,
            )
            return bool(self.incomplete_functions)

    def synth(self) -> STerm:
        with RecordTime("Synthesis"):
            synthesizer = make_synthesizer(self.cfg.synthesizer)
            mapping: dict[Name, Term] = synthesize_holes(
                self.typing_ctx,
                self.evaluation_ctx,
                self.core,
                self.incomplete_functions,
                self.metadata,
                synthesizer,
                self.cfg.synthesis_budget,
                self.cfg.synthesis_ui,
            )

            core_ast_anf: Term = self.core
            for k, v in mapping.items():
                if v is not None:
                    core_ast_anf = substitution(core_ast_anf, v, k)

            sterm_mapping: dict[Name, STerm] = {k: lift(v) for k, v in mapping.items() if v is not None}

            self.cfg.synthesis_ui.display_results(core_ast_anf, sterm_mapping, self.cfg.synthesis_format)

            return lift(core_ast_anf)

    def pretty_print(self, filename: str = None, should_be_fixed: bool = False) -> None:
        aeon_code = read_file(filename)
        prog: Program = parse_main_program(aeon_code, filename=filename)
        prog = bind_program(prog, [])
        pretty_prog = pretty_print_node(prog)
        print(pretty_prog)
        print()
        if should_be_fixed:
            try:
                with open(filename, "w") as f:
                    f.write(pretty_prog)
                print(f"successfully reformatted {filename}")
            except IOError as _:
                print(f"error formatting file {filename}", file=sys.stderr)



from dataclasses import dataclass
from enum import IntEnum, Enum, auto
from functools import reduce
from typing import Callable, List, Tuple

from aeon.sugar.ast_helpers import true, false
from aeon.sugar.program import (
    SLet,
    SIf,
    SAbstraction,
    SApplication,
    SAnnotation,
    SRec,
    STypeAbstraction,
    STypeApplication,
    SLiteral,
    SVar,
    SHole,
    STerm,
    Node,
    ImportAe,
    TypeDecl,
    Decorator,
    InductiveDecl,
    Definition,
    Program,
)
from aeon.sugar.stypes import SType, STypeVar, SRefinedType, SAbstractionType, STypePolymorphism, STypeConstructor
from aeon.utils.name import Name
from aeon.utils.pprint_helpers import (
    Doc,
    parens,
    group,
    concat,
    line,
    text,
    nest,
    DEFAULT_WIDTH,
    soft_line,
    DEFAULT_TAB_SIZE,
    insert_between,
    nil,
    hard_line,
)


class Operation(Enum):
    INFIX_MULTIPLICATIVE = auto()
    INFIX_ADDITIVE = auto()
    INFIX_RELATIONAL = auto()
    INFIX_AND = auto()
    INFIX_OR = auto()
    POLYMORPHISM = auto()
    LET = auto()
    IF = auto()
    ARROW = auto()
    LAMBDA = auto()
    REFINED_TYPE = auto()
    ANNOTATION = auto()
    TYPE_CONSTRUCTOR = auto()
    APPLICATION = auto()
    TYPE_APPLICATION = auto()
    LITERAL = auto()


class Precedence(IntEnum):
    POLYMORPHISM = 1
    LET = 2
    IF = 2
    ANNOTATION = 3
    ARROW = 4
    LAMBDA = 5
    REFINED_TYPE = 6
    TYPE_CONSTRUCTOR = 7
    INFIX_OR = 8
    INFIX_AND = 9
    INFIX_RELATIONAL = 10
    INFIX_ADDITIVE = 11
    INFIX_MULTIPLICATIVE = 12
    APPLICATION = 13
    TYPE_APPLICATION = 14
    LITERAL = 15


class Associativity(Enum):
    LEFT = auto()
    RIGHT = auto()
    NONE = auto()


@dataclass(frozen=True)
class OperationInfo:
    precedence: Precedence
    associativity: Associativity


OPERATION_INFO = {
    Operation.POLYMORPHISM: OperationInfo(Precedence.POLYMORPHISM, Associativity.NONE),
    Operation.LET: OperationInfo(Precedence.LET, Associativity.NONE),
    Operation.IF: OperationInfo(Precedence.IF, Associativity.NONE),
    Operation.ANNOTATION: OperationInfo(Precedence.ANNOTATION, Associativity.NONE),
    Operation.ARROW: OperationInfo(Precedence.ARROW, Associativity.RIGHT),
    Operation.LAMBDA: OperationInfo(Precedence.LAMBDA, Associativity.RIGHT),
    Operation.REFINED_TYPE: OperationInfo(Precedence.REFINED_TYPE, Associativity.NONE),
    Operation.TYPE_CONSTRUCTOR: OperationInfo(Precedence.TYPE_CONSTRUCTOR, Associativity.NONE),
    Operation.INFIX_MULTIPLICATIVE: OperationInfo(Precedence.INFIX_MULTIPLICATIVE, Associativity.LEFT),
    Operation.INFIX_ADDITIVE: OperationInfo(Precedence.INFIX_ADDITIVE, Associativity.LEFT),
    Operation.INFIX_RELATIONAL: OperationInfo(Precedence.INFIX_RELATIONAL, Associativity.NONE),
    Operation.INFIX_AND: OperationInfo(Precedence.INFIX_AND, Associativity.LEFT),
    Operation.INFIX_OR: OperationInfo(Precedence.INFIX_OR, Associativity.LEFT),
    Operation.APPLICATION: OperationInfo(Precedence.APPLICATION, Associativity.LEFT),
    Operation.TYPE_APPLICATION: OperationInfo(Precedence.TYPE_APPLICATION, Associativity.LEFT),
    Operation.LITERAL: OperationInfo(Precedence.LITERAL, Associativity.NONE),
}


def get_operation_precedence(operation: Operation) -> Precedence:
    return OPERATION_INFO[operation].precedence


def get_operation_associativity(operation: Operation) -> Associativity:
    return OPERATION_INFO[operation].associativity


INFIX_MULTIPLICATIVE = {"*", "/", "%", "*.", "/.", "%."}
INFIX_ADDITIVE = {"+", "-", "+.", "-."}
INFIX_RELATIONAL = {"<", "<=", ">", ">=", "==", "!="}
INFIX_AND = {"&&"}
INFIX_OR = {"||"}

AEON_INFIX_OPERATORS = INFIX_MULTIPLICATIVE | INFIX_ADDITIVE | INFIX_RELATIONAL | INFIX_AND | INFIX_OR


def get_sterm_operation(sterm: STerm) -> Operation:
    match sterm:
        case SLet():
            return Operation.LET
        case SIf():
            return Operation.IF
        case SAbstraction():
            return Operation.LAMBDA
        case SApplication(fun=SApplication(fun=SVar(name=op_name), arg=_, loc=_), arg=_, loc=_) if (
            op_name.pretty() in AEON_INFIX_OPERATORS
        ):
            return get_infix_op(op_name)
        case SApplication():
            return Operation.APPLICATION
        case SAnnotation():
            return Operation.ANNOTATION
        case SRec():
            return Operation.LET
        case STypeAbstraction():
            return Operation.POLYMORPHISM
        case STypeApplication():
            return Operation.TYPE_APPLICATION
        case SLiteral() | SVar() | SHole():
            return Operation.LITERAL
        case _:
            return Operation.LITERAL


def get_infix_op(op_name: Name) -> Operation:
    op = op_name.pretty()
    if op in INFIX_MULTIPLICATIVE:
        return Operation.INFIX_MULTIPLICATIVE
    elif op in INFIX_ADDITIVE:
        return Operation.INFIX_ADDITIVE
    elif op in INFIX_RELATIONAL:
        return Operation.INFIX_RELATIONAL
    elif op in INFIX_AND:
        return Operation.INFIX_AND
    elif op in INFIX_OR:
        return Operation.INFIX_OR
    else:
        return Operation.INFIX_OR


def get_stype_operation(stype: SType) -> Operation:
    match stype:
        case STypeVar():
            return Operation.LITERAL
        case SRefinedType():
            return Operation.REFINED_TYPE
        case SAbstractionType():
            return Operation.ARROW
        case STypePolymorphism():
            return Operation.POLYMORPHISM
        case STypeConstructor():
            return Operation.TYPE_CONSTRUCTOR
        case _:
            return Operation.LITERAL


class Side(Enum):
    LEFT = "left"
    RIGHT = "right"
    NONE = "none"


@dataclass(frozen=True)
class ParenthesisContext:
    parent_precedence: Precedence
    child_side: Side


def needs_parens(child_operation: Operation, parenthesis_context: ParenthesisContext) -> bool:
    child_precedence = get_operation_precedence(child_operation)
    parent_precedence = parenthesis_context.parent_precedence
    child_side = parenthesis_context.child_side
    child_associativity = get_operation_associativity(child_operation)

    return needs_parens_aux(child_associativity, child_precedence, child_side, parent_precedence)


def needs_parens_aux(
    child_associativity: Associativity, child_precedence: Precedence, child_side: Side, parent_precedence: Precedence
) -> bool:
    if child_precedence < parent_precedence:
        return True
    if child_precedence > parent_precedence:
        return False

    if child_associativity == Associativity.NONE:
        return False
    if child_associativity == Associativity.LEFT and child_side == Side.RIGHT:
        return True
    if child_associativity == Associativity.RIGHT and child_side == Side.LEFT:
        return True

    return False


def add_parens_if_needed(doc: Doc, child_operation: Operation, parenthesis_context: ParenthesisContext) -> Doc:
    return parens(doc) if needs_parens(child_operation, parenthesis_context) else doc


def pretty_stype_with_parens(stype: SType, parenthesis_context: ParenthesisContext) -> Doc:
    child_operation = get_stype_operation(stype)
    child_pretty = stype_pretty(stype, ParenthesisContext(get_operation_precedence(child_operation), Side.NONE))
    return add_parens_if_needed(child_pretty, child_operation, parenthesis_context)


def pretty_sterm_with_parens(sterm: STerm, parenthesis_context: ParenthesisContext, depth: int) -> Doc:
    child_op = get_sterm_operation(sterm)
    child_pretty = sterm_pretty(sterm, ParenthesisContext(get_operation_precedence(child_op), Side.NONE), depth)
    return add_parens_if_needed(child_pretty, child_op, parenthesis_context)


def format_infix_application(left: STerm, right: STerm, op_name: Name, depth: int) -> Doc:
    infix_op = get_infix_op(op_name)
    infix_precedence = get_operation_precedence(infix_op)
    pretty_left = pretty_sterm_with_parens(left, ParenthesisContext(infix_precedence, Side.LEFT), depth + 1)
    pretty_right = pretty_sterm_with_parens(right, ParenthesisContext(infix_precedence, Side.RIGHT), depth + 1)
    return group(concat([pretty_left, line(), text(op_name.pretty()), line(), pretty_right]))


def pretty_param_doc(param_name: Name, param_type: SType) -> Doc:
    match param_type:
        case SRefinedType(name=ref_name, type=ref_type, refinement=refinement):
            if ref_name.pretty() == param_name.pretty():
                return parens(
                    concat(
                        [
                            text(ref_name.pretty()),
                            text(" : "),
                            stype_pretty(ref_type),
                            text(" | "),
                            sterm_pretty(refinement, ParenthesisContext(Precedence.REFINED_TYPE, Side.RIGHT), depth=1),
                        ]
                    )
                )
    return parens(
        concat(
            [
                text(param_name.pretty()),
                text(" : "),
                stype_pretty(param_type),
            ]
        )
    )


def pretty_print_function_definition(
    func_name_doc: Doc, params: List[Tuple[Name, SType]], return_type: SType
) -> Tuple[Doc, int]:
    first_param_name, first_param_type = params[0]
    first_param_doc = pretty_param_doc(first_param_name, first_param_type)
    func_header = concat([text("def "), func_name_doc, text(" "), first_param_doc])

    indent_after_first_param = len("def ") + len(func_name_doc.layout(0)) + 1
    additional_params = nil()
    for param_name, param_type in params[1:]:
        additional_params = concat(
            [
                additional_params,
                line(),
                pretty_param_doc(param_name, param_type),
            ]
        )
    full_func_def = concat(
        [
            func_header,
            nest(indent_after_first_param, concat([additional_params, line(), text(": "), stype_pretty(return_type)])),
        ]
    )
    return full_func_def, indent_after_first_param


def unwrap_abstraction_types(stype: SType) -> Tuple[List[Tuple[Name, SType]], SType]:
    named_vars = []
    curr = stype
    while True:
        match curr:
            case SAbstractionType(var_name=_var_name, var_type=_var_type, type=next_type):
                named_vars.append((_var_name, _var_type))
                curr = next_type
            case other:
                return named_vars, other


def strip_matching_abstractions(abstraction: STerm, arguments: List[Tuple[Name, SType]]) -> STerm:
    stripped_abstraction = abstraction
    for argument_name, _ in arguments:
        match stripped_abstraction:
            case SAbstraction(var_name=_var_name, body=body) if _var_name.name == argument_name.name:
                stripped_abstraction = body
            case _:
                break
    return stripped_abstraction


def stype_pretty(stype: SType, context: ParenthesisContext = None) -> Doc:
    if context is None:
        context = ParenthesisContext(parent_precedence=Precedence.LITERAL, child_side=Side.NONE)

    match stype:
        case STypeVar(name=name):
            return text(f"{name.pretty()}")

        case SRefinedType(name=name, type=type_, refinement=refinement):
            pretty_name = text(name.pretty())

            type_doc = pretty_stype_with_parens(type_, ParenthesisContext(Precedence.REFINED_TYPE, Side.RIGHT))
            refinement_doc = pretty_sterm_with_parens(
                refinement, ParenthesisContext(Precedence.REFINED_TYPE, Side.RIGHT), 1
            )

            var_type_pair = concat([pretty_name, text(" : "), type_doc])

            return group(
                concat(
                    [
                        text("{"),
                        nest(DEFAULT_WIDTH, concat([soft_line(), var_type_pair, text(" | "), refinement_doc])),
                        soft_line(),
                        text("}"),
                    ]
                )
            )

        case SAbstractionType(var_name=var_name, var_type=var_type, type=_type):
            pretty_var_name = text(var_name.pretty())
            var_type_doc = pretty_stype_with_parens(var_type, ParenthesisContext(Precedence.ANNOTATION, Side.RIGHT))

            left_doc = concat([pretty_var_name, text(" : "), var_type_doc])
            left_doc = add_parens_if_needed(
                left_doc, Operation.ANNOTATION, ParenthesisContext(Precedence.ARROW, Side.LEFT)
            )

            right_doc = pretty_stype_with_parens(_type, ParenthesisContext(Precedence.ARROW, Side.RIGHT))

            return group(concat([left_doc, text(" ->"), line(), right_doc]))

        case STypePolymorphism(name=name, kind=kind, body=body):
            pretty_name = text(name.pretty())
            pretty_kind = text(str(kind))  # should be changed to skind pretty in the future

            pretty_body = pretty_stype_with_parens(body, ParenthesisContext(Precedence.POLYMORPHISM, Side.RIGHT))

            left = concat([text("∀"), pretty_name, text(" : "), pretty_kind])

            return group(concat([left, text(" ."), nest(DEFAULT_TAB_SIZE, concat([line(), pretty_body]))]))

        case STypeConstructor(name=name, args=args):
            pretty_name = text(name.pretty())
            if not args:
                return pretty_name

            pretty_args = [
                pretty_stype_with_parens(arg, ParenthesisContext(Precedence.TYPE_CONSTRUCTOR, Side.RIGHT))
                for arg in args
            ]

            pretty_args_doc = insert_between(text(" "), pretty_args)
            return group(concat([pretty_name, text(" "), pretty_args_doc]))

        case _:
            return text(str(stype))


def sterm_pretty(sterm: STerm, context: ParenthesisContext = None, depth: int = 0) -> Doc:
    if context is None:
        context = ParenthesisContext(parent_precedence=Precedence.LITERAL, child_side=Side.NONE)

    match sterm:
        case SLiteral(value=value, type=_type):
            if _type == STypeConstructor(Name("String", 0)):
                return text(f'"{value}"')
            return text(str(value))

        case SVar(name=name):
            return text(name.pretty())

        case SHole(name=name):
            return text("?" + name.pretty())

        case SAnnotation(expr=expr, type=_type):
            expr_doc = pretty_sterm_with_parens(expr, ParenthesisContext(Precedence.ANNOTATION, Side.LEFT), depth + 1)
            type_doc = stype_pretty(_type, ParenthesisContext(Precedence.ANNOTATION, Side.RIGHT))
            return group(concat([expr_doc, text(" :"), nest(DEFAULT_TAB_SIZE, concat([line(), type_doc]))]))

        case SIf(cond=cond, then=then, otherwise=otherwise):
            pretty_cond = pretty_sterm_with_parens(cond, ParenthesisContext(Precedence.IF, Side.NONE), depth + 1)
            pretty_then = pretty_sterm_with_parens(then, ParenthesisContext(Precedence.IF, Side.NONE), depth + 1)
            pretty_otherwise = pretty_sterm_with_parens(
                otherwise, ParenthesisContext(Precedence.IF, Side.NONE), depth + 1
            )

            return group(
                concat(
                    [
                        text("if "),
                        pretty_cond,
                        line(),
                        text("then "),
                        nest(DEFAULT_TAB_SIZE, pretty_then),
                        line(),
                        text("else "),
                        nest(DEFAULT_TAB_SIZE, pretty_otherwise),
                    ]
                )
            )

        case SApplication(fun=fun, arg=arg) as term:
            match term:
                case (
                    SApplication(fun=SApplication(fun=SVar(name=op_name), arg=left, loc=_), arg=right, loc=_)
                    | SApplication(
                        fun=SApplication(fun=STypeApplication(body=SVar(name=op_name), type=_), arg=left, loc=_),
                        arg=right,
                        loc=_,
                    )
                ) if op_name.pretty() in AEON_INFIX_OPERATORS:
                    return format_infix_application(left, right, op_name, depth)

            if depth == 0:
                if isinstance(fun, SVar) and fun.name.pretty() == "main":
                    return nil()

            pretty_fun = pretty_sterm_with_parens(fun, ParenthesisContext(Precedence.APPLICATION, Side.LEFT), depth + 1)
            pretty_arg = pretty_sterm_with_parens(
                arg, ParenthesisContext(Precedence.APPLICATION, Side.RIGHT), depth + 1
            )

            return group(concat([pretty_fun, line(), pretty_arg]))

        case SAbstraction(var_name=var_name, body=body):
            pretty_var_name = text(var_name.pretty())
            _left = concat([text("\\"), pretty_var_name, text(" ->")])

            _pretty_body = pretty_sterm_with_parens(body, ParenthesisContext(Precedence.ARROW, Side.RIGHT), depth + 1)

            return group(concat([_left, nest(DEFAULT_TAB_SIZE, concat([line(), _pretty_body]))]))

        case SLet(var_name=var_name, var_value=var_value, body=body):
            pretty_var_name = text(var_name.pretty())
            pretty_var_value = pretty_sterm_with_parens(
                var_value, ParenthesisContext(Precedence.LET, Side.NONE), depth + 1
            )
            pretty_body = pretty_sterm_with_parens(body, ParenthesisContext(Precedence.LET, Side.NONE), depth)

            if depth == 0:
                return group(
                    concat(
                        [
                            text("def "),
                            pretty_var_name,
                            text(" ="),
                            nest(DEFAULT_TAB_SIZE, concat([line(), pretty_var_value])),
                            hard_line(),
                            hard_line(),
                            pretty_body,
                        ]
                    )
                )
            else:
                binding = concat([pretty_var_name, text(" = "), pretty_var_value])
                return group(concat([text("let "), binding, text(" in"), line(), pretty_body]))

        case SRec(var_name=var_name, var_type=var_type, var_value=var_value, body=body):  # refazer rec
            pretty_var_name = text(var_name.pretty())
            pretty_var_type = pretty_stype_with_parens(var_type, ParenthesisContext(Precedence.ANNOTATION, Side.RIGHT))
            pretty_var_value = pretty_sterm_with_parens(
                var_value, ParenthesisContext(Precedence.LET, Side.NONE), depth + 1
            )
            pretty_body = pretty_sterm_with_parens(body, ParenthesisContext(Precedence.LET, Side.NONE), depth)

            if depth != 0:
                pretty_binding = concat([pretty_var_name, text(" : "), pretty_var_type, text(" = "), pretty_var_value])
                return group(concat([text("let "), pretty_binding, text(" in"), line(), pretty_body]))

            match var_type:
                case SAbstractionType() as _type:
                    named_vars, final_type = unwrap_abstraction_types(_type)

                    pretty_func_definition, alignment = pretty_print_function_definition(
                        pretty_var_name, named_vars, final_type
                    )

                    var_value = strip_matching_abstractions(var_value, named_vars)

                    pretty_var_value = pretty_sterm_with_parens(
                        var_value, ParenthesisContext(Precedence.LET, Side.NONE), depth + 1
                    )

                    full_type = group(
                        concat(
                            [
                                pretty_func_definition,
                                text(" {"),
                                nest(DEFAULT_TAB_SIZE, concat([line(), pretty_var_value])),
                                line(),
                                text("}"),
                            ]
                        )
                    )

                case _:
                    def_line = concat([text("def "), pretty_var_name, text(" : ")])
                    full_type = group(
                        concat(
                            [
                                def_line,
                                nest(len("def ") + len(var_name.pretty()) + DEFAULT_TAB_SIZE, pretty_var_type),
                                line(),
                                text("= "),
                                nest(DEFAULT_TAB_SIZE, pretty_var_value),
                            ]
                        )
                    )
            return group(concat([full_type, hard_line(), hard_line(), pretty_body]))

        case STypeAbstraction(name=name, kind=kind, body=body):
            pretty_name = text(name.pretty())
            pretty_kind = text(str(kind))  # should be changed to skind pretty in the future
            pretty_body = pretty_sterm_with_parens(
                body, ParenthesisContext(Precedence.APPLICATION, Side.RIGHT), depth + 1
            )

            pretty_kind_def = concat([pretty_name, text(" : "), pretty_kind])
            pretty_binding = concat([pretty_kind_def, text("."), pretty_body])
            return group(concat([text("ƛ"), pretty_binding]))

        case STypeApplication(body=body, type=_type):
            pretty_body = sterm_pretty(body, ParenthesisContext(Precedence.APPLICATION, Side.LEFT), depth + 1)
            return group(pretty_body)

        case _:
            return text(str(sterm))


def normalize_term(term: STerm, context: dict[Name, STerm] = None, seen: set[Name] = None) -> STerm:
    if context is None:
        context = {}
    if seen is None:  # seen is used as a fix to infinite recursion
        seen = set()
    match term:
        case SLiteral(value=value, type=type):
            return SLiteral(value=value, type=type)
        case SVar(name=name):
            if name in seen or name not in context:
                return SVar(name=name)
            simplified_term = context[name]
            new_seen = seen.copy()
            new_seen.add(name)
            return normalize_term(simplified_term, context, new_seen)
        case SHole(name=name):
            return SHole(name=name)
        case SAnnotation(expr=expr, type=type):
            simplified_expr = normalize_term(expr, context, seen)
            return SAnnotation(expr=simplified_expr, type=type)
        case SIf(cond=cond, then=then, otherwise=otherwise):
            simplified_cond = normalize_term(cond, context, seen)
            if simplified_cond == true:
                simplified_then = normalize_term(then, context, seen)
                return simplified_then
            elif simplified_cond == false:
                simplified_otherwise = normalize_term(otherwise, context, seen)
                return simplified_otherwise
            else:
                simplified_then = normalize_term(then, context, seen)
                simplified_otherwise = normalize_term(otherwise, context, seen)
                return SIf(cond=simplified_cond, then=simplified_then, otherwise=simplified_otherwise)
        case SApplication(fun=fun, arg=arg):
            simplified_fun = normalize_term(fun, context, seen)
            simplified_arg = normalize_term(arg, context, seen)

            if isinstance(simplified_fun, SAbstraction):
                new_context = context.copy()
                new_context[simplified_fun.var_name] = simplified_arg
                return normalize_term(simplified_fun.body, new_context, seen)
            else:
                return SApplication(fun=simplified_fun, arg=simplified_arg)
        case SAbstraction(var_name=var_name, body=body):
            simplified_body = normalize_term(body, context, seen)
            return SAbstraction(var_name=var_name, body=simplified_body)
        case SLet(var_name=var_name, var_value=var_value, body=body):
            if var_name.pretty() == "anf":
                context_copy = context.copy()
                context_copy[var_name] = normalize_term(var_value, context, seen)
                simplified_body = normalize_term(body, context_copy, seen)
                return simplified_body
            match body:
                case SHole(name=name):
                    return SLet(var_name=name, var_value=normalize_term(var_value, context, seen), body=SVar(name=name))
                case _:
                    return SLet(
                        var_name=var_name,
                        var_value=normalize_term(var_value, context, seen),
                        body=normalize_term(body, context, seen),
                    )
        case SRec(var_name=var_name, var_type=var_type, var_value=var_value, body=body):
            if var_name.pretty() == "anf":
                context_copy = context.copy()
                context_copy[var_name] = normalize_term(var_value, context, seen)
                simplified_body = normalize_term(body, context_copy, seen)
                return simplified_body
            match body:
                case SHole(name=name):
                    return SRec(
                        var_name=name,
                        var_type=var_type,
                        var_value=normalize_term(var_value, context, seen),
                        body=SVar(name=name),
                    )
                case _:
                    return SRec(
                        var_name=var_name,
                        var_type=var_type,
                        var_value=normalize_term(var_value, context, seen),
                        body=normalize_term(body, context, seen),
                    )
        case STypeAbstraction(name=name, kind=kind, body=body):
            simplified_body = normalize_term(body, context, seen)
            return STypeAbstraction(name=name, kind=kind, body=simplified_body)
        case _:
            return term


def rename_unused_variables(term: STerm):
    # TODO
    return term


def pretty_print_sterm(term: STerm) -> str:
    simplified_term = simplify_sterm(term)
    return str(sterm_pretty(simplified_term))


def simplify_sterm(term: STerm) -> STerm:
    def apply_func(acc: STerm, func: Callable[[STerm], STerm]) -> STerm:
        return func(acc)

    functions: List[Callable[[STerm], STerm]] = [normalize_term, rename_unused_variables]
    return reduce(apply_func, functions, term)


def node_pretty(node: Node) -> Doc:
    match node:
        case ImportAe(path=path, func=func):
            return (
                group(concat([text("import"), line(), text(f'"{path}";')]))
                if not func
                else group(
                    concat([text("import"), line(), text(func), line(), text("from"), line(), text(f'"{path}";')])
                )
            )
        case TypeDecl(name=name):
            return group(concat([text("type"), line(), text(f"{name.pretty()};")]))
        case Decorator(name=name, macro_args=macro_args):
            pretty_macro_args = [sterm_pretty(simplify_sterm(arg)) for arg in macro_args]
            pretty_macro_args_doc = parens(insert_between(concat([text(","), line()]), pretty_macro_args))

            return concat([text("@"), text(name.pretty()), pretty_macro_args_doc])
        case InductiveDecl(name=name, args=args, constructors=constructors, measures=measures):
            pretty_name = concat([text("inductive"), line(), text(name.pretty())])
            pretty_args = group(insert_between(line(), [text(arg.pretty()) for arg in args]))
            pretty_constructors = group(
                insert_between(line(), [concat([text("| "), node_pretty(cons)]) for cons in constructors])
            )
            pretty_measures = group(
                insert_between(line(), [concat([text("+ "), node_pretty(meas)]) for meas in measures])
            )

            return group(insert_between(line(), [pretty_name, pretty_args, pretty_constructors, pretty_measures]))
        case Definition(name=name, foralls=foralls, args=args, type=type_, body=body, decorators=decorators):
            for var_name, _ in reversed(args):
                body = SAbstraction(var_name=var_name, body=body)

            for var_name, var_type in reversed(args):
                type_ = SAbstractionType(var_name=var_name, var_type=var_type, type=type_)

            for type_name, kind in reversed(foralls):
                body = STypeAbstraction(name=type_name, kind=kind, body=body)

            pretty_decorators = insert_between(line(), [node_pretty(decorator) for decorator in decorators])

            match type_:
                case SAbstractionType() as _type:
                    named_vars, final_type = unwrap_abstraction_types(_type)

                    pretty_func_definition, alignment = pretty_print_function_definition(
                        text(name.pretty()), named_vars, final_type
                    )

                    stripped_body = strip_matching_abstractions(body, named_vars)

                    pretty_body = pretty_sterm_with_parens(
                        stripped_body, ParenthesisContext(Precedence.LET, Side.NONE), depth=1
                    )

                    definition_doc = group(
                        concat(
                            [
                                pretty_func_definition,
                                text(" {"),
                                nest(DEFAULT_TAB_SIZE, concat([line(), pretty_body])),
                                line(),
                                text("}"),
                            ]
                        )
                    )

                case _:
                    pretty_var_name = name.pretty()
                    pretty_var_type = stype_pretty(type_)
                    pretty_body = sterm_pretty(body, depth=1)

                    def_line = concat([text("def "), text(pretty_var_name), text(" : ")])

                    definition_doc = group(
                        concat(
                            [
                                def_line,
                                nest(len("def ") + len(pretty_var_name) + DEFAULT_TAB_SIZE, pretty_var_type),
                                line(),
                                text("= "),
                                nest(DEFAULT_TAB_SIZE, pretty_body),
                                text(";"),
                            ]
                        )
                    )

            return group(concat([pretty_decorators, hard_line() if decorators else nil(), definition_doc]))

        case Program(imports=imports, type_decls=type_decls, inductive_decls=inductive_decls, definitions=definitions):
            pretty_blocks = [
                group(insert_between(line(), [node_pretty(_import) for _import in imports])) if imports else nil(),
                group(insert_between(line(), [node_pretty(type_decl) for type_decl in type_decls]))
                if type_decls
                else nil(),
                group(insert_between(line(), [node_pretty(inductive_decl) for inductive_decl in inductive_decls]))
                if inductive_decls
                else nil(),
                group(insert_between(concat([hard_line(), hard_line()]), [node_pretty(defn) for defn in definitions]))
                if definitions
                else nil(),
            ]

            return group(
                insert_between(concat([hard_line(), hard_line()]), [block for block in pretty_blocks if block != nil()])
            )
    return nil()


def pretty_print_node(node: Node) -> str:
    return str(node_pretty(node))



from __future__ import annotations

from typing import Any

from aeon.core.terms import Abstraction, TypeAbstraction, TypeApplication
from aeon.core.terms import Annotation
from aeon.core.terms import Application
from aeon.core.terms import Hole
from aeon.core.terms import If
from aeon.core.terms import Let
from aeon.core.terms import Literal
from aeon.core.terms import Rec
from aeon.core.terms import Term
from aeon.core.terms import Var
from aeon.utils.name import Name

real_eval = eval


class EvaluationContext:
    variables: dict[Name, Any]

    def __init__(self, prev: dict[Name, Any] | None = None):
        if prev:
            self.variables = {k: v for (k, v) in prev.items()}
        else:
            self.variables = {}

    def with_var(self, name: Name, value: Any):
        assert isinstance(name, Name)
        v = self.variables.copy()
        v.update({name: value})
        return EvaluationContext(v)

    def get(self, name: Name):
        return self.variables[name]


def is_native_var(fun: Any):
    return fun == real_eval


def is_native_import(fun: Term):
    match fun:
        case TypeApplication(t, _):
            return is_native_import(t)
        case Var(Name("native_import", _)):
            return True
        case _:
            return False


# pattern match term
def eval(t: Term, ctx: EvaluationContext = EvaluationContext()) -> Any:
    match t:
        case Literal(value, _):
            return value
        case Var(name):
            return ctx.get(name)
        case Abstraction(var_name, body):
            return lambda k: eval(body, ctx.with_var(var_name, k))
        case Application(fun, arg):
            f = eval(fun, ctx)
            argv = eval(arg, ctx)
            if is_native_var(f):
                assert isinstance(argv, str)

                python_ctx = {str(name): v for name, v in globals().items()}
                python_ctx.update({str(name.name): v for name, v in ctx.variables.items()})
                e = real_eval(argv, python_ctx)
            else:
                e = f(argv)
            if is_native_import(fun):
                globals()[argv] = e
            return e
        case Let(var_name, var_value, body):
            return eval(body, ctx.with_var(var_name, eval(var_value, ctx)))
        case Rec(var_name, _, var_value, body):
            if isinstance(var_value, Abstraction):
                fun = var_value

                def v(x):
                    return eval(
                        fun.body,
                        ctx.with_var(var_name, v).with_var(fun.var_name, x),
                    )

            else:
                v = eval(var_value, ctx)
            return eval(t.body, ctx.with_var(t.var_name, v))
        case If(cond, then, otherwise):
            c = eval(cond, ctx)
            if c:
                return eval(then, ctx)
            else:
                return eval(otherwise, ctx)
        case Annotation(expr, _):
            return eval(expr, ctx)
        case Hole(name):
            args = ", ".join([str(n.name) for n in ctx.variables])
            print(f"Context ({args})")
            h = input(f"Enter value for hole {t} in Python: ")
            return real_eval(h, {str(name): v for name, v in ctx.variables.items()})

        case TypeAbstraction(_, _, body):
            return eval(body, ctx)
        case TypeApplication(body, _):
            return eval(body, ctx)
        case _:
            assert False, f"Unknown case {t}"



from __future__ import annotations


def curried(x, argc=None):
    if argc is None:
        argc = x.__code__.co_argcount

    def p(*a):
        if len(a) == argc:
            return x(*a)

        def q(*b):
            return x(*(a + b))

        return curried(q, argc - len(a))

    return p



from __future__ import annotations

from dataclasses import dataclass
from functools import reduce
from typing import Any
from typing import Generator
from loguru import logger

from z3 import Function
from z3 import Int
from z3 import Solver
from z3 import sat
from z3 import unknown
from z3.z3 import And
from z3.z3 import Bool
from z3.z3 import BoolRef
from z3.z3 import BoolSort
from z3.z3 import Const
from z3.z3 import DeclareSort
from z3.z3 import Float64
from z3.z3 import Implies
from z3.z3 import IntSort
from z3.z3 import Not
from z3.z3 import Or
from z3.z3 import String
from z3.z3 import StringSort
from z3.z3 import SortRef
from z3.z3types import Z3Exception

from aeon.core.liquid import LiquidApp
from aeon.core.types import LiquidHornApplication, TypeConstructor
from aeon.core.liquid import LiquidLiteralBool
from aeon.core.liquid import LiquidLiteralFloat
from aeon.core.liquid import LiquidLiteralInt
from aeon.core.liquid import LiquidLiteralString
from aeon.core.liquid import LiquidTerm
from aeon.core.liquid import LiquidVar
from aeon.core.liquid_ops import mk_liquid_and
from aeon.core.substitutions import substitution_in_liquid
from aeon.core.types import AbstractionType, RefinedType, Top, TypePolymorphism
from aeon.core.types import Type
from aeon.core.types import TypeVar
from aeon.core.types import t_bool, t_int, t_float, t_string, t_unit
from aeon.verification.vcs import Conjunction
from aeon.verification.vcs import Constraint
from aeon.verification.vcs import Implication
from aeon.verification.vcs import LiquidConstraint
from aeon.verification.vcs import UninterpretedFunctionDeclaration
from aeon.utils.name import Name, fresh_counter

smt_function_types: dict[str, list[Type]] = {
    "smtEqInt": [t_int, t_int, t_bool],
    "smtEqFloat": [t_float, t_float, t_bool],
    "smtEqBool": [t_float, t_float, t_bool],
    "smtEqString": [t_string, t_string, t_bool],
    "smtNeqInt": [t_int, t_int, t_bool],
    "smtNeqFloat": [t_float, t_float, t_bool],
    "smtNeqBool": [t_float, t_float, t_bool],
    "smtNeqString": [t_string, t_string, t_bool],
    "smtNot": [t_bool, t_bool],
    "smtLeqInt": [t_int, t_int, t_bool],
    "smtLeqFloat": [t_float, t_float, t_bool],
    "smtLtInt": [t_int, t_int, t_bool],
    "smtLtFloat": [t_float, t_float, t_bool],
    "smtAnd": [t_bool, t_bool, t_bool],
    "smtOr": [t_bool, t_bool, t_bool],
    "smtPlusInt": [t_int, t_int, t_int],
    "smtMinusInt": [t_int, t_int, t_int],
    "smtMultInt": [t_int, t_int, t_int],
    "smtDivInt": [t_int, t_int, t_int],
    "smtModInt": [t_int, t_int, t_int],
    "smtPlusFloat": [t_float, t_float, t_float],
    "smtMinusFloat": [t_float, t_float, t_float],
    "smtMultFloat": [t_float, t_float, t_float],
    "smtDivFloat": [t_float, t_float, t_float],
    "smtImplies": [t_bool, t_bool, t_bool],
}

smt_function_translation: dict[str, list[str]] = {
    "==": ["smtEqBool", "smtEqInt", "smtEqFloat", "smtEqString"],
    "!=": ["smtNeqBool", "smtNeqInt", "smtNeqFloat", "smtNeqString"],
    "!": ["smtNot"],
    "<": ["smtLeqInt", "smtLeqFloat"],
    "<=": ["smtLtInt", "smtLtFloat"],
    "&&": ["smtAnd"],
    "||": ["smtOr"],
    "+": ["smtPlusInt", "smtPlusFloat"],
    "-": ["smtMinusInt", "smtMinusFloat"],
    "*": ["smtMultInt", "smtMultFloat"],
    "/": ["smtDivInt", "smtDivFloat"],
    "%": ["smtModInt"],
    "-->": ["smtImplies"],
}

base_functions: dict[str, Any] = {
    "==": lambda x, y: x == y,
    "!=": lambda x, y: x != y,
    "<": lambda x, y: x < y,
    "<=": lambda x, y: x <= y,
    ">": lambda x, y: x > y,
    ">=": lambda x, y: x >= y,
    "!": lambda x: Not(x),
    "&&": lambda x, y: And(x, y),
    "||": lambda x, y: Or(x, y),
    "+": lambda x, y: x + y,
    "-": lambda x, y: x - y,
    "*": lambda x, y: x * y,
    "/": lambda x, y: x / y,
    "%": lambda x, y: x % y,
    "-->": lambda x, y: Implies(x, y),
}


@dataclass
class SMTContext:
    sorts: list[str]
    functions: dict[str, AbstractionType]
    variables: dict[str, TypeConstructor]
    premises: list[LiquidTerm]

    def with_sort(self, name: Name) -> SMTContext:
        return SMTContext(self.sorts + [str(name)], self.functions, self.variables, self.premises)

    def with_function(self, name: Name, ty: AbstractionType) -> SMTContext:
        return SMTContext(self.sorts, {**self.functions, str(name): ty}, self.variables, self.premises)

    def with_var(self, name: Name, ty: TypeConstructor) -> SMTContext:
        return SMTContext(self.sorts, self.functions, {**self.variables, str(name): ty}, self.premises)

    def with_premise(self, p: LiquidTerm) -> SMTContext:
        return SMTContext(self.sorts, self.functions, self.variables, self.premises + [p])


@dataclass(init=False)
class CanonicConstraint:
    """Represents SMT-valid constraints."""

    sorts: list[str]
    functions: dict[str, AbstractionType]
    variables: dict[str, TypeConstructor]
    premise: LiquidTerm
    conclusion: LiquidTerm

    def __init__(self, ctx: SMTContext, pos: LiquidTerm):
        self.sorts = ctx.sorts
        self.functions = ctx.functions
        self.variables = ctx.variables
        self.premise = reduce(mk_liquid_and, ctx.premises, LiquidLiteralBool(True))
        self.conclusion = pos


def rename_constraint(c: Constraint, old_name: Name, new_name: Name) -> Constraint:
    """Renames a binder within the constraint, to make it is unique."""
    match c:
        case LiquidConstraint(expr):
            nexpr = substitution_in_liquid(expr, LiquidVar(new_name), old_name)
            return LiquidConstraint(expr=nexpr)
        case Conjunction(c1, c2):
            return Conjunction(rename_constraint(c1, old_name, new_name), rename_constraint(c2, old_name, new_name))
        case Implication(name, base, pred, seq):
            if name == new_name:
                return c
            else:
                npred = substitution_in_liquid(pred, LiquidVar(new_name), old_name)
                nseq = rename_constraint(seq, old_name, new_name)
                return Implication(name, base, npred, nseq)
        case UninterpretedFunctionDeclaration(name, absty, seq):
            nseq = rename_constraint(seq, old_name, new_name)
            return UninterpretedFunctionDeclaration(name, absty, nseq)
        case _:
            assert False, f"Unexpected case {c} ({type(c)})"


def flatten(c: Constraint, ctx: SMTContext | None = None) -> Generator[CanonicConstraint]:
    """Flattens a constraint into a list of SMT-valid constraints."""
    if ctx is None:
        ctx = SMTContext(["Top"], {}, {}, [])
    match c:
        case LiquidConstraint(expr):
            yield CanonicConstraint(ctx, expr)
        case Conjunction(c1, c2):
            yield from flatten(c1, ctx)
            yield from flatten(c2, ctx)
        case Implication(oname, base, pred, seq):
            name = Name(oname.name, fresh_counter.fresh())
            pred = substitution_in_liquid(pred, LiquidVar(name), oname)
            seq = rename_constraint(seq, oname, name)
            match base:
                case TypeVar(iname):
                    base = TypeConstructor(iname)
                case TypeConstructor(iname, []):
                    pass
                case TypeConstructor(iname, args):
                    mangle_name = str(iname) + "_" + "_".join(str(a) for a in args)
                    nname = Name(mangle_name, fresh_counter.fresh())
                    base = TypeConstructor(nname)
                case _:
                    assert False, f"{base} ({type(base)}) is not a base type."
            yield from flatten(seq, ctx.with_var(name, base).with_premise(pred))
        case UninterpretedFunctionDeclaration(name, ty, seq):
            assert isinstance(c, UninterpretedFunctionDeclaration)
            yield from flatten(seq, ctx.with_function(name, ty))
        case _:
            assert False, f"Cannot flatten {c}."


s = Solver()
(s.set(timeout=200),)


def smt_valid(constraint: Constraint) -> bool:
    """Verifies if a constraint is true using Z3."""
    n = 0
    for c in flatten(constraint):
        n += 1
        s.push()

        # TODO now: Add monomorphic, uncurried functions here
        try:
            smt_c = translate(c)
        except ZeroDivisionError:
            continue
        if smt_c is False:
            continue
        s.add(smt_c)
        result = s.check()
        s.pop()
        if result == sat:
            return False
        elif result == unknown:
            return False

    return True


def type_of_variable(variables: list[tuple[str, Any]], name: str) -> Any:
    for na, ref in reversed(variables):
        if na == name:
            return ref
    vars = ", ".join([x[0] for x in variables])
    logger.error(f"No variable {name} in the context: {vars}")
    assert False


sort_cache: dict[str, SortRef] = {}


def mk_vars(variables: dict[str, TypeConstructor], sorts: dict[str, SortRef]) -> dict[str, Any]:
    return {name: make_variable(name, base) for name, base in variables.items()}


def get_sort(base: Type) -> SortRef:
    match base:
        case Top() | TypeConstructor(Name("Top", _)):
            return DeclareSort("Top")
        case TypeConstructor(Name("Int", _)):
            return IntSort()
        case TypeConstructor(Name("Bool", _)):
            return BoolSort()
        case TypeConstructor(Name("Float", _)):
            return Float64()
        case TypeConstructor(Name("String", _)):
            return StringSort()
        case TypeConstructor(name, _):
            return IntSort()
            # This will be reenable once we have typeclasses
            # if name not in sort_cache:
            #     sort_cache[name] = DeclareSort(name)
            # return sort_cache[name]
        case TypeVar(name):
            assert False, f"TypeVar {name} should not be used in SMT solver."
        case _:
            raise Exception(f"SMT sort of {base} not implemented.")


def unrefine_type(base: Type):
    """Removes refinements from type."""
    match base:
        case RefinedType(_, ty, _):
            return ty
        case AbstractionType(name, aty, rty):
            return AbstractionType(name, unrefine_type(aty), unrefine_type(rty))
        case TypePolymorphism(name, kind, body):
            return TypePolymorphism(name, kind, unrefine_type(body))
        case TypeConstructor(name, args):
            return TypeConstructor(name, [unrefine_type(a) for a in args])

        case _:
            return base


def uncurry(base: AbstractionType) -> tuple[list[TypeConstructor], TypeConstructor]:
    current: Type = unrefine_type(base)
    inputs = []
    vars_to_remove = []

    while isinstance(current, TypePolymorphism):
        vars_to_remove.append(current.name)
        current = current.body

    while isinstance(current, AbstractionType):
        match current.var_type:
            case TypeConstructor(_, []):
                inputs.append(current.var_type)
            case TypeConstructor(_, _):
                inputs.append(t_int)
            case Top():
                inputs.append(t_unit)
            case TypeVar(name):
                if name in vars_to_remove:
                    inputs.append(t_int)
                else:
                    inputs.append(TypeConstructor(name))
            case _:
                assert False, f"Unknown SMT type {current.var_type} in {base}."
        current = current.type

    if isinstance(current, Top):
        current = t_unit
    assert isinstance(current, TypeConstructor), f"Unknown SMT type {current} in {base}."
    return (inputs, current)


def make_variable(name: str, base: TypeConstructor | AbstractionType | Top) -> Any:
    match base:
        case Top():
            return Const(name, get_sort(base))
        case TypeConstructor(Name("Int", _)):
            return Int(name)
        case TypeConstructor(Name("Bool", _)):
            return Bool(name)
        case TypeConstructor(Name("Float", _), _):
            import z3

            v = z3.Real(name)
            return v
            # TODO: see problem with version below
            # fpsort = FPSort(8, 24)
            # return FP(name, fpsort)
        case TypeConstructor(Name("String", _)):
            return String(name)
        case TypeConstructor(_, _):
            return Int(name)
            # TODO: we always use int, in the case of a typevar.
            # return Const(name, get_sort(base))
        case TypeVar(_):
            return Int(name)
        case AbstractionType(_, _, _):
            if name in base_functions:
                return base_functions[name]
            else:
                input_types, output_type = uncurry(base)
                args = [get_sort(x) for x in input_types] + [get_sort(output_type)]
                return Function(name, *args)
        case _:
            assert False, f"No var: {name}, with base {base}."


def translate_liq(t: LiquidTerm, variables: dict[str, Any]):
    match t:
        case LiquidLiteralBool(b):
            return b
        case LiquidLiteralInt(i):
            return i
        case LiquidLiteralFloat(f):
            return f
        case LiquidLiteralString(s):
            return s
        case LiquidVar(name):
            return variables[str(name)]
        case LiquidHornApplication(name, args):
            assert False, "LiquidHornApplication should not get to SMT solver!"
        case LiquidApp(fun_name, args):
            fun = base_functions.get(fun_name.name, variables.get(str(fun_name), None))
            assert fun is not None, f"Function {fun_name} not found." + str(variables)
            args = [translate_liq(a, variables) for a in args]
            try:
                return fun(*args)
            except Z3Exception as e:
                raise e

        case _:
            assert False, f"Cannot translate {t}."


def mk_sorts(sorts: list[str]) -> dict[str, SortRef]:
    return {name: get_sort(TypeConstructor(Name(name, 0))) for name in sorts}


def mk_funs(functions: dict[str, AbstractionType], sorts: dict[str, SortRef]) -> dict[str, Any]:
    funs = {}
    for name, ty in functions.items():
        input_types, output_type = uncurry(ty)
        args = [sorts.get(str(x), get_sort(x)) for x in input_types] + [
            sorts.get(str(output_type), get_sort(output_type))
        ]
        funs[name] = Function(name, *args)
    return funs


def translate(
    c: CanonicConstraint,
) -> BoolRef | bool:
    sorts = mk_sorts(c.sorts)
    functions = mk_funs(c.functions, sorts)
    variables = mk_vars(c.variables, sorts)
    e1 = translate_liq(c.premise, variables | functions)
    e2 = translate_liq(c.conclusion, variables | functions)
    if isinstance(e2, bool) and e2 is True:
        return False
    if isinstance(e1, bool) and isinstance(e2, bool):
        return e1 and not e2
    return And(e1, Not(e2))
